from django.conf.urls import url, include

from driver27 import views

urlpatterns = [
    url(r'^$', views.race_list, name='race-list'),
    url(r'^(?P<race_id>\d+)$', views.race_view, name='race-view'),
    url(r'^(?P<race_id>\d+)/(?P<rank>[-\w\d]+)$', views.race_view, name='race-rank'),
]



# -*- coding: utf-8 -*-
from django.test import TestCase
from .common import CommonResultTestCase
from ..models import Result, Seat, TeamSeason, ContenderSeason, SeatPeriod, CompetitionTeam
from ..records import get_record_config
from ..punctuation import get_punctuation_config
from django.core.exceptions import ValidationError

from collections import namedtuple


class ResultTestCase(TestCase, CommonResultTestCase):
    def _insert_massive_results(self, seat, season, results):
        for new_result in results:
            cur_race = self.get_test_race(season=season, round=new_result['round'])
            del new_result['round']
            self.get_test_result(race=cur_race, seat=seat, **new_result)

    def test_result_shorcuts(self):
        seat = self.get_test_seat()
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition)
        race = self.get_test_race(season=season)
        self.get_test_competition_team(competition=competition, team=seat.team)
        result = self.get_test_result(seat=seat, race=race)
        self.assertEquals(result.driver, result.seat.driver)
        self.assertEquals(result.team, result.seat.team)

    def test_result_points(self):
        seat = self.get_test_seat()
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition)
        race = self.get_test_race(season=season)
        self.get_test_competition_team(competition=competition, team=seat.team)
        result = self.get_test_result(seat=seat, race=race)

        result.qualifying = 2
        result.finish = 2
        self.assertIsNone(result.save())

        # result.finish = 2 = ?? points
        self.assertGreater(result.points, 0)
        team_season, created = TeamSeason.objects.get_or_create(team=seat.team, season=season)
        self.assertGreater(team_season.get_points(), 0)
        race_points = result.points

        # change scoring to add fastest_lap point
        race.fastest_car = result.seat
        self.assertIsNone(race.save())

        # modify scoring to check if fastest_lap scoring is counted
        punctuation_config = season.get_punctuation_config()
        punctuation_config['fastest_lap'] = 1
        self.assertGreater(result.points_calculator(punctuation_config), race_points)
        self.assertEquals(race.fastest, result.seat)

    def test_result_str(self):
        seat = self.get_test_seat()
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition)
        race = self.get_test_race(season=season)
        self.get_test_competition_team(competition=competition, team=seat.team)
        result = self.get_test_result(race=race, seat=seat)
        result.finish = 14
        result_str = '{seat} ({race})'.format(seat=result.seat, race=result.race)
        result_str_checkered_flag = result_str + ' - {finish}º'.format(finish=result.finish)
        self.assertEqual(str(result), result_str_checkered_flag)

        result_str_out = result_str + ' - DNF'
        result.finish = None
        self.assertEqual(str(result), result_str_out)

    def test_result_seat_exception(self):
        self.assertRaises(ValidationError, self.get_test_result)

    def test_streak_results(self):
        record_config = get_record_config('PODIUM').get('filter')
        seat = self.get_test_seat()
        driver = seat.driver
        competition_1 = self.get_test_competition()
        season_1 = self.get_test_season(competition=competition_1, rounds=8)
        self.get_test_competition_team(competition=competition_1, team=seat.team)

        results_1 = [
            {'round': 1, 'finish': 3, 'qualifying': 4},
            {'round': 2, 'finish': 2, 'qualifying': 13},
            {'round': 3, 'finish': 1, 'qualifying': 2},
            {'round': 4, 'finish': 1, 'qualifying': 1},
            {'round': 5, 'finish': 1, 'qualifying': 1},
            {'round': 6, 'finish': 8, 'qualifying': 5},
            {'round': 7, 'finish': 3, 'qualifying': 4},
            {'round': 8, 'finish': 3, 'qualifying': 3},
        ]

        self._insert_massive_results(seat=seat, season=season_1, results=results_1)

        self.assertEqual(driver.get_streak(**record_config), 2)
        self.assertEqual(driver.get_streak(max_streak=True, **record_config), 5)

        self.assertEqual(driver.get_points(season=season_1), 142)
        self.assertEqual(season_1.leader['points'], driver.get_points(season=season_1))
        self.assertTrue(season_1.has_champion())

        # season 2
        # same competition
        # streak continues
        # streak in season and streak global are diff
        year_2 = season_1.year+1
        season_2 = self.get_test_season(competition=competition_1, year=year_2)

        results_2 = [
            {'round': 1, 'finish': 3, 'qualifying': 4},
            {'round': 2, 'finish': 2, 'qualifying': 11},
            {'round': 3, 'finish': 1, 'qualifying': 20},
            {'round': 4, 'finish': 1, 'qualifying': 12}
        ]

        self._insert_massive_results(seat=seat, season=season_2, results=results_2)

        self.assertEqual(driver.get_streak(**record_config), 6)
        self.assertEqual(driver.get_season(season_2).get_streak(**record_config), 4)

        self.assertEqual(driver.get_points(season=season_2), 83)
        self.assertEqual(season_2.leader['points'], driver.get_points(season=season_2))

        # season 3
        # diff competition
        # streak continues
        # streak in season and streak in competition are the same but not streak global
        year_3 = season_2.year+1
        competition_2 = self.get_test_competition_2()
        self.get_test_competition_team(competition=competition_2, team=seat.team)
        season_3 = self.get_test_season(competition=competition_2, year=year_3)

        results_3 = [
            {'round': 1, 'finish': 1, 'qualifying': 2},
            {'round': 2, 'finish': 2, 'qualifying': 2},
            {'round': 3, 'finish': 3, 'qualifying': 2},
            {'round': 4, 'finish': 1, 'qualifying': 1},
            {'round': 5, 'finish': 1, 'qualifying': 1}
        ]

        self._insert_massive_results(seat=seat, season=season_3, results=results_3)

        self.assertEqual(driver.get_streak(**record_config), 11)
        self.assertEqual(driver.get_streak(result_filter={'competition': competition_2}, **record_config), 5)
        self.assertEqual(driver.get_season(season_3).get_streak(**record_config), 5)

        self.assertEqual(driver.get_points(season=season_3), 108)
        self.assertEqual(season_3.leader['points'], driver.get_points(season=season_3))


        # season 4
        # back to competition 1
        # streak global continues
        # but streak in competition continues from season 2 (skip season 3 - other competition)
        year_4 = season_3.year+1
        season_4 = self.get_test_season(competition=competition_1, year=year_4)

        results_4 = [
            {'round': 1, 'finish': 1, 'qualifying': 2},
            {'round': 2, 'finish': 1, 'qualifying': 2},
        ]

        self._insert_massive_results(seat=seat, season=season_4, results=results_4)

        self.assertEqual(driver.get_streak(**record_config), 13)
        self.assertEqual(driver.get_streak(result_filter={'competition': competition_1}, **record_config), 8)
        self.assertEqual(driver.get_season(season_4).get_streak(**record_config), 2)

        # bye streak
        # max_streak is past_streak and curr streak is 1

        results_bye_streak = [
            {'round': 3, 'finish': 5, 'qualifying': 3},
            {'round': 4, 'finish': 1, 'qualifying': 2},
        ]

        self._insert_massive_results(seat=seat, season=season_4, results=results_bye_streak)

        self.assertEqual(driver.get_streak(**record_config), 1)
        self.assertEqual(driver.get_streak(result_filter={'competition': competition_1}, **record_config), 1)
        self.assertEqual(driver.get_season(season_4).get_streak(**record_config), 1)
        self.assertEqual(driver.get_streak(max_streak=True, **record_config), 13)
        self.assertEqual(driver.get_streak(max_streak=True, result_filter={'competition': competition_1},
                                           **record_config), 8)
        self.assertEqual(driver.get_season(season_4).get_streak(max_streak=True, **record_config), 2)
        # and records in competition_2/season_3 are not altered
        self.assertEqual(driver.get_streak(result_filter={'competition': competition_2}, **record_config), 5)
        self.assertEqual(driver.get_season(season_3).get_streak(**record_config), 5)

        # check records
        self.assertEqual(driver.get_stats(**record_config), 19, 'Global record')
        self.assertEqual(driver.get_stats(competition=competition_1, **record_config), 14, 'Record in comp_1')
        self.assertEqual(driver.get_stats(competition=competition_2, **record_config), 5, 'Record in comp_2')
        self.assertEqual(driver.get_stats(season=season_1, **record_config), 7, 'Record in season_1')
        self.assertEqual(driver.get_stats(season=season_2, **record_config), 4, 'Record in season_2')
        self.assertEqual(driver.get_stats(season=season_3, **record_config), 5, 'Record in season_3')
        self.assertEqual(driver.get_stats(season=season_4, **record_config), 3, 'Record in season_4')

        self.assertTrue(driver.is_active)

        punctuation_config = get_punctuation_config('F1-10+6')

        self.assertEqual(driver.get_points(), 418)
        self.assertEqual(driver.get_points(punctuation_config=punctuation_config), 150)

        self.assertEqual(driver.get_points(season=season_4), 85)
        self.assertEqual(season_4.leader['points'], driver.get_points(season=season_4))

    def test_rank(self):
        seat_a = self.get_test_seat()
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition)
        race = self.get_test_race(season=season)
        self.get_test_competition_team(competition=competition, team=seat_a.team)
        result_a = self.get_test_result(seat=seat_a, race=race, qualifying=2, finish=1)

        seat_b = self.get_test_seat_teammate(seat_a=seat_a)
        self.get_test_result(seat=seat_b, race=race, qualifying=1, finish=3)

        # pole
        self.assertEquals(race.pole, seat_b)
        # winner
        self.assertEquals(race.winner, seat_a)

        # season contender = 2
        self.assertEquals(season.drivers.count(), 2)
        self.assertEquals(len(season.points_rank()), 2)
        self.assertEquals(len(season.olympic_rank()), 2)
        self.assertEquals(len(season.team_points_rank()), 1) # seat a and seat b is in the same team
        self.assertEquals(season.leader['driver'], seat_a.driver) # seat a.driver is the winner, the leader
        self.assertEquals(season.team_leader['team'], seat_a.team)
        self.assertEqual(season.runner_up['driver'], seat_b.driver)

        # contenderseason.get_points
        driver_a = seat_a.driver
        contender_season = ContenderSeason(driver=driver_a, season=season)
        self.assertGreater(contender_season.get_points(), 0)
        self.assertGreater(contender_season.get_points(1), 0)

    def test_team_competition_validation(self):
        seat_a = self.get_test_seat()
        competition = self.get_test_competition()
        season_year = 2018
        season = self.get_test_season(competition=competition, year=season_year)
        race = self.get_test_race(season=season, round=1)
        self.assertRaises(ValidationError, self.get_test_result, **{'seat': seat_a, 'race': race,
                                                                    'qualifying': 2, 'finish': 2})

    def test_period_limitation(self):
        period_year = 2017
        seat_a = self.get_test_seat()
        self.assertTrue(SeatPeriod.objects.create(seat=seat_a, from_year=period_year, until_year=period_year))

        competition = self.get_test_competition()
        season_year = 2018
        season = self.get_test_season(competition=competition, year=season_year)
        race = self.get_test_race(season=season, round=1)
        self.assertRaises(ValidationError, self.get_test_result, **{'seat': seat_a, 'race': race,
                                                                    'qualifying': 2, 'finish': 2})

    def test_rank_integrity(self):
        seat_a = self.get_test_seat()
        competition_a = self.get_test_competition()
        season_a = self.get_test_season(competition=competition_a, year=2018)
        race_season_a = self.get_test_race(season=season_a, round=1)
        self.get_test_competition_team(competition=competition_a, team=seat_a.team)

        result_a = self.get_test_result(seat=seat_a, race=race_season_a, qualifying=1, finish=1)
        self.assertEqual(season_a.leader['points'], 25)

        season_b = self.get_test_season(competition=competition_a, year=2019)
        race_season_b = self.get_test_race(season=season_b, round=1)
        result_b = self.get_test_result(seat=seat_a, race=race_season_b, qualifying=1, finish=1)

        rank_b = season_b.points_rank()
        self.assertEqual(season_b.leader['points'], 25)

        rank_competition_a = competition_a.points_rank()
        leader_comp_a = rank_competition_a[0]
        self.assertEqual(leader_comp_a['points'], 50)

        competition_c = self.get_test_competition_2()
        season_c = self.get_test_season(competition=competition_c, year=2020)
        race_season_c = self.get_test_race(season=season_c, round=1)
        self.get_test_competition_team(competition=competition_c, team=seat_a.team)
        result_c = self.get_test_result(seat=seat_a, race=race_season_c, qualifying=1, finish=1)

        rank_c = season_c.points_rank()
        leader_c = rank_c[0]
        self.assertEqual(leader_c['points'], 25)

        rank_competition_c = competition_c.points_rank()
        leader_comp_c = rank_competition_c[0]
        self.assertEqual(leader_comp_c['points'], 25)

    def test_seat_team_exception(self):
        seat = self.get_test_seat()
        race = self.get_test_race()
        result_args = {'seat': seat, 'race': race, 'qualifying': 1, 'finish': 3}
        self.assertRaises(ValidationError, self.get_test_result, **result_args)

    def test_duplicate_driver(self):
        seat_a = self.get_test_seat()
        competition = self.get_test_competition()
        self.assertTrue(self.get_test_competition_team(competition=competition, team=seat_a.team))
        season = self.get_test_season(competition=competition)
        race = self.get_test_race(season=season)
        result_a_args = {'seat': seat_a, 'race': race, 'qualifying': 1, 'finish': 3}
        result_a = self.get_test_result(**result_a_args)

        seat_a_other_team = self.get_test_seat_same_driver_other_team(seat_a)
        result_b_args = {'seat': seat_a_other_team, 'race': race, 'qualifying': 2, 'finish': 2}
        self.assertRaises(ValidationError, self.get_test_result, **result_b_args)
        self.assertEqual(race.results.count(), 1)

        # test swapfield. Seat_a loses the pole, Seat_b is poleman
        seat_b = self.get_test_seat_teammate(seat_a)
        self.assertTrue(self.get_test_result(seat=seat_b, race=race, qualifying=1, finish=3))
        self.assertEqual(race.results.count(), 2)
        self.assertEqual(race.pole, seat_b)

    def test_has_champions_half_punctuation(self):
        # create two players
        seat_a = self.get_test_seat()
        seat_b = self.get_test_seat_teammate(seat_a)

        # create the season with 3 rounds, the last with half punctuation
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition, rounds=3)
        self.get_test_competition_team(competition=competition, team=seat_a.team)

        # create 3 races
        race_1 = self.get_test_race(season=season, round=1)
        race_2 = self.get_test_race(season=season, round=2)
        race_3 = self.get_test_race(season=season, round=3, alter_punctuation='half')

        # race 1
        # seat_a: 25, seat_b: 18
        self.get_test_result(seat=seat_a, race=race_1, qualifying=2, finish=1)
        self.get_test_result(seat=seat_b, race=race_1, qualifying=1, finish=2)
        self.assertFalse(season.has_champion())

        # race 2
        # seat_a: 50, seat_b: 36
        # with regular punctuation, race 3 will determine the champion
        # but with half punctuation, distance of 14 points with 12.5 on the line,
        # seat_a is the champion with a pending race
        self.get_test_result(seat=seat_a, race=race_2, qualifying=1, finish=1)
        self.get_test_result(seat=seat_b, race=race_2, qualifying=4, finish=2)
        self.assertEqual(season.pending_points(), 12.5)
        self.assertTrue(season.has_champion())

    def test_has_champions_double_punctuation(self):
        # create two players
        seat_a = self.get_test_seat()
        seat_b = self.get_test_seat_teammate(seat_a)

        # create the season with 3 rounds, the last with double punctuation
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition, rounds=3)
        self.get_test_competition_team(competition=competition, team=seat_a.team)

        # create 3 races
        race_1 = self.get_test_race(season=season, round=1)
        race_2 = self.get_test_race(season=season, round=2)
        race_3 = self.get_test_race(season=season, round=3, alter_punctuation='double')

        # race 1
        # seat_a: 25, seat_b: 0
        self.get_test_result(seat=seat_a, race=race_1, qualifying=2, finish=1)
        self.get_test_result(seat=seat_b, race=race_1, qualifying=1, finish=None, retired=True)
        self.assertFalse(season.has_champion())

        # race 2
        # seat_a: 43, seat_b: 15
        # with regular punctuation, seat_a would already be the champion
        # but with double punctuation, seat_b a disadvantage of 28 points with 50 pending.
        self.get_test_result(seat=seat_a, race=race_2, qualifying=1, finish=2)
        self.get_test_result(seat=seat_b, race=race_2, qualifying=4, finish=3)
        self.assertEqual(season.pending_points(), 50)
        self.assertFalse(season.has_champion())

        # race 3
        # seat_a: 63, seat_b: 65 (with double punctuation, seat_b is the champion)
        # seat_a: 53, seat_b: 40 (with regular punctuation, seat_a is the champion)
        self.get_test_result(seat=seat_a, race=race_3, qualifying=8, finish=5)
        self.get_test_result(seat=seat_b, race=race_3, qualifying=3, finish=1)
        self.assertTrue(season.the_champion()['driver'], seat_b.driver)




# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from .common import retro_encode, CommonTeamSeasonTestCase
from ..models import TeamSeason, CompetitionTeam

class TeamSeasonTestCase(TestCase, CommonTeamSeasonTestCase):
    def test_team_season_validation(self):
        team = self.get_test_team()
        competition_b = self.get_test_competition_2()
        self.assertTrue(CompetitionTeam.objects.create(competition=competition_b, team=team))

    def test_team_season_unicode(self):

        team = self.get_test_team()
        competition_a = self.get_test_competition()
        season = self.get_test_season(competition=competition_a)
        self.assertTrue(CompetitionTeam.objects.create(competition=competition_a, team=team))
        team_season_args = {'team': team, 'season': season}
        self.assertTrue(TeamSeason.objects.create(**team_season_args))
        team_season = TeamSeason.objects.get(**team_season_args)
        # STR team_season
        expected_team_season = u'{team} in {season}'.format(team=team_season.team, season=season)
        self.assertEquals(str(team_season), retro_encode(expected_team_season))
        team_season.sponsor_name = 'Sponsored Team'
        self.assertIsNone(team_season.save())
        expected_team_season = u'{sponsor} in {season}'.format(sponsor=team_season.sponsor_name, season=season)
        self.assertEquals(str(team_season), retro_encode(expected_team_season))


# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand, CommandError
from driver27.models import Driver, Team, Seat

import sys
if sys.version_info < (3, 0):
    try:
        import unicodecsv as csv
    except ImportError:
        import csv
else:
    import csv


class Command(BaseCommand):
    help = 'Export seats to csv'

    def get_config(self, export_attr):
        if export_attr == 'drivers':
            fieldnames = ['id', 'first_name', 'last_name', 'country', 'year_of_birth']
            export_cls = Driver
        elif export_attr == 'teams':
            fieldnames = ['id', 'name', 'full_name', 'country']
            export_cls = Team
        else:
            fieldnames = ['id', 'driver_id', 'driver__last_name', 'driver__first_name', 'team_id', 'team__name']
            export_cls = Seat
        objects = list(export_cls.objects.values(*fieldnames))
        return {'fieldnames': fieldnames, 'objects': objects}

    def add_arguments(self, parser):
        parser.add_argument('csv',)
        parser.add_argument(
            '--export',
            default='seats',
            help='By default, export seats. Options: seats, drivers, teams',
        )

    def handle(self, *args, **options):

        with open(options['csv'], 'wb') as csvfile:
            export_config = self.get_config(options['export'])
            writer = csv.DictWriter(csvfile, fieldnames=export_config['fieldnames'])
            writer.writeheader()

            for entry in export_config['objects']:
                writer.writerow(entry)



# -*- coding: utf-8 -*-
import sys
import six
from django.core.exceptions import ValidationError
from ..models import Driver, Competition, Team, Season, Circuit
from ..models import TeamSeason, Seat, GrandPrix, Race, Result, ContenderSeason, CompetitionTeam


def retro_encode(text):
    if sys.version_info < (3, 0):
        try:
            return text.encode('utf-8')
        except UnicodeDecodeError:
            return text
    else:
        return text


class CommonCreateTestCase(object):
    def set_test_create(self, model, **kwargs):
        return model.objects.create(**kwargs)


class CommonDriverTestCase(CommonCreateTestCase):
    def get_test_driver(self, **kwargs):
        defaults = {
            'last_name': u'García',
            'first_name': u'Juan'
        }
        defaults.update(**kwargs)
        return self.set_test_create(model=Driver, **defaults)

    def get_test_driver_2(self):
        kwargs = {
            'last_name': u'López',
            'first_name': u'Jaime'
        }
        return self.get_test_driver(**kwargs)


class CommonCompetitionTestCase(CommonCreateTestCase):
    def get_test_competition(self, **kwargs):
        defaults = {
            'name': u'Competición A',
            'full_name': u'Competición ABC'
        }
        defaults.update(**kwargs)
        return self.set_test_create(model=Competition, **defaults)

    def get_test_competition_2(self):
        kwargs = {
            'name': u'Competición B',
            'full_name': u'Competición BDF'
        }
        return self.get_test_competition(**kwargs)


class CommonTeamTestCase(CommonCompetitionTestCase):
    def get_test_team(self, **kwargs):
        defaults = {
            'name': u'Escudería Tec Auto',
            'full_name': u'Escudería Tec Auto'
        }
        defaults.update(**kwargs)
        return self.set_test_create(model=Team, **defaults)

    def get_test_team_2(self):
        kwargs = {
            'name': u'D27R',
            'full_name': u'Driver 27 Racing Team'
        }
        return self.get_test_team(**kwargs)

    def get_test_competition_team(self, **kwargs):
        defaults = {}
        defaults.update(**kwargs)
        if 'competition' not in defaults:
            defaults['competition'] = self.get_test_competition()
        if 'team' not in defaults:
            defaults['team'] = self.get_test_team()
        return self.set_test_create(model=CompetitionTeam, **defaults)


class CommonSeasonTestCase(CommonCompetitionTestCase):
    def get_test_season(self, **kwargs):
        defaults = {
            'year': 2016,
            'punctuation': u'F1-25'
        }
        defaults.update(**kwargs)
        if 'competition' not in defaults:
            defaults['competition'] = self.get_test_competition()
        return self.set_test_create(model=Season, **defaults)


class CommonTeamSeasonTestCase(CommonTeamTestCase, CommonSeasonTestCase):
    pass


class CommonSeatTestCase(CommonDriverTestCase, CommonTeamTestCase):
    def get_test_seat(self, **kwargs):
        defaults = {}
        defaults.update(**kwargs)
        if 'driver' not in defaults:
            defaults['driver'] = self.get_test_driver()
        if 'team' not in defaults:
            defaults['team'] = self.get_test_team()
        return self.set_test_create(model=Seat, **defaults)

    def get_test_seat_teammate(self, seat_a):
        driver = self.get_test_driver_2()
        team = seat_a.team
        seat_args = {'driver': driver, 'team': team}
        return self.get_test_seat(**seat_args)

    def get_test_seat_same_driver_other_team(self, seat_a):
        driver = seat_a.driver
        team = self.get_test_team_2()
        seat_args = {'driver': driver, 'team': team}
        return self.get_test_seat(**seat_args)


class CommonRaceTestCase(CommonSeasonTestCase):
    def get_test_race(self, **kwargs):
        defaults = {'round': 1}
        defaults.update(**kwargs)
        if 'season' not in defaults:
            defaults['season'] = self.get_test_season()
        return self.set_test_create(model=Race, **defaults)

    def get_test_circuit(self, **kwargs):
        defaults = {
            'name': u'Autódromo de Jacarepaguá',
            'city': u'Jacarepaguá',
            'opened_in': 1978
        }
        defaults.update(**kwargs)
        return self.set_test_create(model=Circuit, **defaults)

    def get_test_grandprix(self, **kwargs):
        defaults = {'name': u'Grande Prêmio do Brasil'}
        defaults.update(**kwargs)
        return self.set_test_create(model=GrandPrix, **defaults)


class CommonResultTestCase(CommonSeatTestCase, CommonRaceTestCase):
    def get_test_result(self, **kwargs):
        defaults = {}
        defaults.update(**kwargs)
        if 'seat' not in defaults:
            defaults['seat'] = self.get_test_seat()
        if 'race' not in defaults:
            defaults['race'] = self.get_test_race()
        return self.set_test_create(model=Result, **defaults)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from collections import namedtuple

from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext as _
from django_countries.fields import CountryField
from slugify import slugify
from swapfield.fields import SwapIntegerField

from . import lr_intr, lr_diff
from .points_calculator import PointsCalculator
from .punctuation import get_punctuation_config
from .rank import AbstractRankModel
from .stats import AbstractStreakModel, AbstractStatsModel, TeamStatsModel, StatsByCompetitionModel, SeasonStatsModel

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from .common import DRIVER27_NAMESPACE

ResultTuple = namedtuple('ResultTuple',
                         'qualifying finish fastest_lap wildcard retired race_id circuit grand_prix country competition year ' \
                         'round alter_punctuation points')


def get_tuples_from_results(results):
    return [get_tuple_from_result(result) for result in results]


def get_tuple_from_result(result, **kwargs):
    result_kwargs = {
        'qualifying': result.qualifying,
        'finish': result.finish,
        'fastest_lap': result.is_fastest,
        'wildcard': result.wildcard,
        'retired': result.retired,
        'race_id': result.race_id,
        'circuit': result.race.circuit,
        'grand_prix': result.race.grand_prix,
        'country': getattr(result.race.grand_prix, 'country', None),
        'competition': result.race.season.competition,
        'year': result.race.season.year,
        'round': result.race.round,
        'alter_punctuation': result.race.alter_punctuation,
        'points': result.points
    }

    result_kwargs.update(**kwargs)

    return ResultTuple(**result_kwargs)


@python_2_unicode_compatible
class Driver(StatsByCompetitionModel):
    """ Main Driver Model. To combine with competitions (Contender) and competition/team (Seat) """
    last_name = models.CharField(max_length=50, verbose_name=_('last name'))
    first_name = models.CharField(max_length=25, verbose_name=_('first name'))
    year_of_birth = models.IntegerField(verbose_name=_('year of birth'), null=True, blank=True)
    country = CountryField(verbose_name=_('country'), null=True, blank=True)
    teams = models.ManyToManyField('Team', through='Seat', related_name='drivers', verbose_name=_('teams'))

    def get_absolute_url(self):
        URL = ':'.join([DRIVER27_NAMESPACE, 'global:driver-profile'])
        return reverse(URL, kwargs = {'driver_id': self.pk})

    @property
    def result_filter_kwargs(self):
        return {}

    @property
    def races(self):
        """ Races with at least one result of this driver """
        return Race.objects.filter(results__seat__driver=self).distinct()

    @property
    def seasons(self):
        """ Season with at least one result of this driver """
        return Season.objects.filter(races__in=self.races.all()).distinct()

    @property
    def competitions(self):
        """ Competition with at least one result of this driver """
        return Competition.objects.filter(seasons__in=self.seasons.all()).distinct()

    @property
    def is_active(self):
        """ A driver is active when has at least one result in last year registered in seasons """
        seasons_ordered_by_desc_year = Season.objects.order_by('-year')
        if seasons_ordered_by_desc_year.count():
            last_year = seasons_ordered_by_desc_year[0].year
            return Result.wizard(driver=self, race__season__year=last_year).count()
        else:
            return True

    def get_summary_points(self, append_to_summary=None, **kwargs):
        kwargs.pop('exclude_position', False)
        punctuation_config = kwargs.pop('punctuation_config', None)
        positions_count_list = self.get_positions_count_list(**kwargs)
        positions_count_str = self.get_positions_count_str(position_list=positions_count_list)
        summary_points = {
            'teams': self.teams_verbose,
            'points': self.get_points(punctuation_config=punctuation_config, **kwargs),
            'pos_list': positions_count_list,
            'pos_str': positions_count_str
        }
        if append_to_summary is not None:
            summary_points.update(**append_to_summary)
        return summary_points

    def _teams_verbose(self, teams):
        return ', '.join([team.name for team in teams])

    @property
    def teams_verbose(self):
        """ A str separated with commas with teams names """
        return self._teams_verbose(self.teams.all())

    def get_results(self, limit_races=None, **extra_filter):
        """ Result of driver """
        return Result.wizard(driver=self, limit_races=limit_races, **extra_filter)

    def get_saved_points(self, limit_races=None, **kwargs):
        """ List of result.points from results of driver """
        results = self.get_results(limit_races=limit_races, **kwargs)
        points = results.values_list('points', flat=True)
        return [point for point in points if point]


    def season_stats_cls(self, season):
        """ Return the instance with driver-season pair """
        return ContenderSeason(driver=self, season=season)

    def get_points_by_season(self, season, **kwargs):
        """
        Return points summary of a driver in a season
        """
        summary_points = self.season_stats_cls(season=season).get_summary_points(**kwargs)
        return summary_points

    def get_points_by_seasons(self, **kwargs):
        """
        Return points summary of each season of driver

        """
        kwargs.pop('season', None)
        seasons = self.seasons.all()
        return [self.get_points_by_season(season=season, **kwargs) for season in seasons]

    def __str__(self):
        return u', '.join((self.last_name, self.first_name))

    class Meta:
        unique_together = ('last_name', 'first_name')
        ordering = ['last_name', 'first_name']
        verbose_name = _('Driver')
        verbose_name_plural = _('Drivers')


@python_2_unicode_compatible
class Competition(AbstractRankModel):
    """ Competition model. Season, races, results... are grouped by competition """
    name = models.CharField(max_length=30, verbose_name=_('competition'), unique=True)
    full_name = models.CharField(max_length=100, unique=True, verbose_name=_('full name'))
    country = CountryField(null=True, blank=True, default=None, verbose_name=_('country'))
    slug = models.SlugField(null=True, blank=True, default=None)

    def get_absolute_url(self):
        URL = ':'.join([DRIVER27_NAMESPACE, 'competition:view'])
        return reverse(URL, kwargs = {'competition_slug': self.slug})

    @property
    def drivers(self):
        """ Queryset of driver with at least one result in a race of this competition """
        return Driver.objects.filter(seats__results__race__season__competition=self).distinct()

    def get_team_rank(self, rank_type, **filters):
        return super(Competition, self).get_team_rank(rank_type=rank_type, competition=self, **filters)

    def get_stats_cls(self, driver):
        return driver

    def get_team_stats_cls(self, team):
        return team

    @property
    def stats_filter_kwargs(self):
        return {'competition': self}

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        super(Competition, self).save(*args, **kwargs)

    def __str__(self):
        return u'{name}'.format(name=self.name)

    class Meta:
        ordering = ['name']
        verbose_name = _('Competition')
        verbose_name_plural = _('Competitions')


class CompetitionTeam(models.Model):
    team = models.ForeignKey('Team', related_name='periods', verbose_name=_('team'))
    competition = models.ForeignKey('Competition', related_name='periods', verbose_name=_('competition'))
    from_year = models.IntegerField(blank=True, null=True, default=None)
    until_year = models.IntegerField(blank=True, null=True, default=None)

    class Meta:
        unique_together = ('team', 'competition', 'from_year', 'until_year')


@python_2_unicode_compatible
class Team(TeamStatsModel, StatsByCompetitionModel):
    """ Team model, unique if is the same in different competition """
    name = models.CharField(max_length=75, verbose_name=_('team'), unique=True)
    full_name = models.CharField(max_length=200, unique=True, verbose_name=_('full name'))
    competitions = models.ManyToManyField('Competition', through='CompetitionTeam', related_name='teams',
                                          verbose_name=_('competitions'))
    country = CountryField(verbose_name=_('country'), blank=True, null=True)

    def get_absolute_url(self):
        URL = ':'.join([DRIVER27_NAMESPACE, 'global:team-profile'])
        return reverse(URL, kwargs = {'team_id': self.pk})

    def _races(self, competition=None, **kwargs):
        """
        Return a races of a team
        With param competition, the filter is different
        """
        kwargs.update(**{'results__seat__team': self})
        if competition is not None:
            kwargs.update(**{'season__competition': competition})
        return Race.objects.filter(**kwargs).distinct()

    @property
    def races(self):
        return self._races().distinct()

    @property
    def seasons(self):
        return Season.objects.filter(races__in=self.races.all()).distinct()

    def get_results(self, competition=None, **extra_filter):
        """ Return results of a team """
        return Result.wizard(team=self, competition=competition, **extra_filter)

    def get_results_by_race(self, race):
        """ Shorcut for return results of a team in a race"""
        return self.get_results(race=race)

    def get_drivers_by_race(self, race):
        """ Shorcut for return drivers of a team in a race"""
        results = self.get_results_by_race(race=race)
        return list(set([result.seat.driver for result in results]))

    def get_drivers_by_race_str(self, race):
        """ Str with drivers name in a race """
        return ' + '.join(sorted(['{driver}'.format(driver=driver) for driver in self.get_drivers_by_race(race)]))


    def get_total_races(self, competition=None, **filters):
        return super(Team, self).get_total_races(competition=competition, **filters)

    def get_doubles_races(self, competition=None, **filters):
        return super(Team, self).get_doubles_races(competition=competition, **filters)

    def get_total_stats(self, competition=None, **filters):
        return super(Team, self).get_total_stats(competition=competition, **filters)

    def season_stats_cls(self, season):
        return TeamSeason.objects.get(season=season, team=self)

    def get_summary_points(self, append_to_summary=None, **kwargs):
        kwargs.pop('exclude_position', False)
        punctuation_config = kwargs.pop('punctuation_config', None)
        positions_count_list = self.get_positions_count_list(**kwargs)
        positions_count_str = self.get_positions_count_str(position_list=positions_count_list)
        summary_points = {
            'points': self.get_points(punctuation_config=punctuation_config, **kwargs),
            'pos_list': positions_count_list,
            'pos_str': positions_count_str
        }

        if append_to_summary is not None:
            summary_points.update(**append_to_summary)
        return summary_points

    def get_points_by_season(self, season, append_team=False, **kwargs):
        summary_points = self.season_stats_cls(season=season).get_summary_points(**kwargs)
        if append_team:
            summary_points['team'] = self
        return summary_points

    def get_points_by_seasons(self, append_team=False, **kwargs):
        kwargs.pop('season', None)
        seasons = self.seasons.all()
        return [self.get_points_by_season(season=season, append_team=append_team, **kwargs) for season in seasons]

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Team')
        verbose_name_plural = _('Teams')


@python_2_unicode_compatible
class SeatPeriod(models.Model):
    seat = models.ForeignKey('Seat', related_name='periods')
    from_year = models.IntegerField(blank=True, null=True, default=None)
    until_year = models.IntegerField(blank=True, null=True, default=None)

    def __str__(self):
        return_text = '{seat}'.format(seat=self.seat)
        if self.from_year:
            return_text += ' from {from_year}'.format(from_year=self.from_year)
        if self.until_year:
            return_text += ' until {until_year}'.format(until_year=self.until_year)
        return return_text


@python_2_unicode_compatible
class Seat(models.Model):
    """ Seat is contender/team relation """
    " If a same contender drives to two teams in same season/competition, he will have many seats as driven teams "
    " e.g. Max Verstappen in 2016 (Seat 1: Toro Rosso, Seat 2: Red Bull) "
    team = models.ForeignKey('Team', related_name='seats', verbose_name=_('team'))
    driver = models.ForeignKey('Driver', related_name='seats', verbose_name=_('driver'), default=None, null=True)

    def __str__(self):
        return _(u'%(driver)s in %(team)s') \
               % {'driver': self.driver, 'team': self.team}

    class Meta:
        unique_together = ('team', 'driver',)
        ordering = ['driver__last_name', 'team']
        verbose_name = _('Seat')
        verbose_name_plural = _('Seats')


@python_2_unicode_compatible
class Circuit(models.Model):
    """ Circuit model. It can be in any competition, grand_prix or season """
    name = models.CharField(max_length=30, verbose_name=_('circuit'), unique=True)
    city = models.CharField(max_length=100, null=True, blank=True, verbose_name=_('city'))
    country = CountryField(verbose_name=_('country'))
    opened_in = models.IntegerField(verbose_name=_('opened in'))

    # @todo Add Clockwise and length
    def __str__(self):
        # @todo Add country name in __str__
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Circuit')
        verbose_name_plural = _('Circuits')


@python_2_unicode_compatible
class GrandPrix(models.Model):
    """ Grand Prix model """
    " default_circuit will be in future the default choice in circuit selector "
    name = models.CharField(max_length=30, verbose_name=_('grand prix'), unique=True)
    country = CountryField(null=True, blank=True, default=None, verbose_name=_('country'))
    first_held = models.IntegerField(null=True, blank=True, verbose_name=_('first held'))
    default_circuit = models.ForeignKey(Circuit, related_name='default_to_grands_prix', null=True,
                                        blank=True, default=None, verbose_name=_('default circuit'))
    competitions = models.ManyToManyField('Competition', related_name='grands_prix', default=None,
                                          verbose_name=_('competitions'))

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Grand Prix')
        verbose_name_plural = _('Grands Prix')

@python_2_unicode_compatible
class Season(AbstractRankModel):
    """ Season model. The main model to restrict races, results and punctuation """
    year = models.IntegerField(verbose_name=_('year'))
    competition = models.ForeignKey(Competition, related_name='seasons', verbose_name=_('competition'))
    rounds = models.IntegerField(blank=True, null=True, default=None, verbose_name=_('rounds'))
    punctuation = models.CharField(max_length=20, null=True, default=None, verbose_name=_('punctuation'))

    def get_params_url(self):
        return {'competition_slug': self.competition.slug, 'year': self.year}

    def get_absolute_url(self):
        URL = ':'.join([DRIVER27_NAMESPACE, 'season:view'])
        return reverse(URL, kwargs=self.get_params_url())

    def get_races_url(self):
        URL = ':'.join([DRIVER27_NAMESPACE, 'season:race-list'])
        return reverse(URL, kwargs=self.get_params_url())


    @property
    def stats_filter_kwargs(self):
        return {'season': self}

    def get_results(self, kwargs=None):
        cache_str = 'season_results_{pk}'.format(pk=self.pk)
        cache_results = cache.get(cache_str)

        if cache_results:
            results = cache_results
        else:
            results = Result.wizard(season=self)
            cache.set(cache_str, results)
        return results

    def get_results_list(self, element='driver', kwargs=None):
        cache_str = 'season_results_{pk}_{element}'.format(pk=self.pk, element=element)
        cache_results_list = cache.get(cache_str)

        if cache_results_list:
            entries = cache_results_list
        else:
            results = self.get_results()
            entries = {}
            for result in results:
                key = getattr(result, element)
                if key not in entries:
                    entries[key] = []
                entries[key].append(result)
            cache.set(cache_str, entries)
        return entries

    def get_results_by_drivers(self):
        return self.get_results_list('driver')


    def get_results_by_teams(self):
        return self.get_results_list('team')

    @property
    def drivers(self):
        """ As season is related with seats, this method is a shorcut to get drivers """
        return Driver.objects.filter(seats__results__race__season=self).distinct()

    @property
    def teams(self):
        """ As season is related with seats, this method is a shorcut to get teams """
        return Team.objects.filter(seats__results__race__season=self).distinct()

    def get_positions_draw(self):
        points_rank = self.points_rank()
        position_draw = []
        for entry in points_rank:
            contender_season = ContenderSeason(driver=entry['driver'], season=self)
            position_draw.append(
                {'pos_list': contender_season.get_positions_in_row(),
                 'driver': entry['driver'],
                 'teams': contender_season.teams_verbose,
                 'pos_str': ''
                 }
            )
        return position_draw

    def _abstract_seats(self, exclude=False):
        seat_filter = {'team__competitions__seasons': self}
        seats = Seat.objects.filter(**seat_filter)
        filter_or_exclude = 'filter' if not exclude else 'exclude'
        return getattr(seats, filter_or_exclude)(

            Q(periods__from_year__lte=self.year) | Q(periods__from_year__isnull=True),
            Q(periods__until_year__gte=self.year) | Q(periods__until_year__isnull=True)
        ).distinct()

    @property
    def seats(self):
        return self._abstract_seats()

    @property
    def no_seats(self):
        return self._abstract_seats(exclude=True)

    def get_stats_cls(self, driver):
        return driver.get_season(self)

    def get_team_stats_cls(self, team):
        team_season, created = TeamSeason.objects.get_or_create(season=self, team=team)
        return team_season

    def get_punctuation_config(self, punctuation_code=None):
        """ Getting the punctuation config. Chosen punctuation will be override temporarily by code kwarg """
        if not punctuation_code:
            punctuation_code = self.punctuation
        return get_punctuation_config(punctuation_code)

    def past_or_pending_races(self, past=True):
        filter_no_results = {'results__pk__isnull': True}
        races = self.races
        if past:
            races = races.exclude(**filter_no_results)
        else:
            races = races.filter(**filter_no_results)
        return races.distinct()

    @property
    def pending_races(self):
        return self.past_or_pending_races(past=False)

    @property
    def past_races(self):
        return self.past_or_pending_races(past=True)

    @property
    def num_pending_races(self):
        expected_pending_races = races_no_results = self.pending_races.count()
        season_rounds = self.rounds
        if season_rounds:
            expected_pending_races = self.rounds - races_no_results
        if races_no_results > expected_pending_races:
            races_no_results = expected_pending_races
        return races_no_results

    def pending_points(self, punctuation_code=None):
        """ Return the maximum of available points taking into account the number of pending races"""
        # @todo fastest_lap and pole points are not counted
        if not punctuation_code:
            punctuation_code = self.punctuation
        punctuation_config = get_punctuation_config(punctuation_code=punctuation_code)
        max_points = 0

        for race in self.pending_races.all():
            # max_result = Result(finish=1, qualifying=1, race=race)
            # @todo fastest_car modify tuple behavior
            max_result = Result(finish=1, qualifying=1, race=race)
            max_tuple = get_tuple_from_result(max_result, fastest_lap=True)
            max_points += PointsCalculator(punctuation_config=punctuation_config).calculator(
                max_tuple
            )
        return max_points

    def leader_window(self, rank=None, punctuation_code=None):
        """ Minimum of current points a contestant must have to have any chance to be a champion.
        This is the subtraction of the points of the leader and the maximum of points that
        can be obtained in the remaining races."""
        punctuation_code_dict = {'punctuation_code': punctuation_code}
        pending_points = self.pending_points(**punctuation_code_dict)
        leader = self.get_leader(rank=rank, **punctuation_code_dict)
        return leader['points'] - pending_points if leader else None

    def only_title_contenders(self, punctuation_code=None):
        """ They are only candidates for the title if they can reach the leader by adding all the pending points."""
        punctuation_code_dict = {'punctuation_code': punctuation_code}
        rank = self.points_rank(**punctuation_code_dict)
        leader_window = self.leader_window(rank=rank, **punctuation_code_dict)

        title_rank = []
        if leader_window is not None:  # Can be zero
            title_rank = [entry for entry in rank if entry['points'] >= leader_window]
        return title_rank

    def the_champion(self, punctuation_code=None):
        title_contenders = self.only_title_contenders(punctuation_code=punctuation_code)
        if len(title_contenders) == 1:
            return title_contenders[0]
        return None

    def has_champion(self, punctuation_code=None):
        """ If only exists one title contender, it has champion """
        return self.the_champion(punctuation_code=punctuation_code) is not None

    def get_leader(self, rank=None, team=False, punctuation_code=None):
        """ Get driver leader or team leader """
        if not rank:
            # if not punctuation_code:
            #     punctuation_code = self.punctuation
            punctuation_code_dict = {'punctuation_code': punctuation_code}
            rank_method = 'team_points_rank' if team else 'points_rank'
            rank = getattr(self, rank_method)(**punctuation_code_dict)
        return rank[0] if len(rank) else None

    @property
    def leader(self):
        """ get_leader(driver) property """
        return self.get_leader()

    @property
    def runner_up(self):
        """ Get the runner_up (driver) """
        rank = self.points_rank()
        return rank[1] if len(rank) > 1 else None

    @property
    def team_leader(self):
        """ get_leader(team) property """
        return self.get_leader(team=True)

    def __str__(self):
        return '{competition}/{year}'.format(competition=self.competition, year=self.year)

    class Meta:
        unique_together = ('year', 'competition')
        ordering = ['year', 'competition__name', ]
        verbose_name = _('Season')
        verbose_name_plural = _('Seasons')


class SeatsSeason(Season):
    """ Aux proxy model to use in Admin for Season/Seat relation from Season """

    class Meta:
        proxy = True
        verbose_name = _('Seats by season')
        verbose_name_plural = _('Seats by season')


@python_2_unicode_compatible
class Race(models.Model):
    """ Race model. It can altered season punctuation (double or half) only in this race"""
    ALTER_PUNCTUATION = (
        ('double', _('Double')),
        ('half', _('Half'))
    )
    season = models.ForeignKey(Season, related_name='races', verbose_name=_('season'))
    round = models.IntegerField(verbose_name=_('round'))
    grand_prix = models.ForeignKey(GrandPrix, related_name='races', blank=True, null=True,
                                   default=None, verbose_name=_('grand prix'))
    circuit = models.ForeignKey(Circuit, related_name='races', blank=True, null=True,
                                default=None, verbose_name=_('circuit'))
    date = models.DateField(blank=True, null=True, default=None, verbose_name=_('date'))
    alter_punctuation = models.CharField(choices=ALTER_PUNCTUATION, null=True, blank=True,
                                         default=None, max_length=6, verbose_name=_('alter punctuation'))
    fastest_car = models.ForeignKey(Seat, related_name='fastest_in', blank=True, null=True, default=None,
                                    on_delete=models.SET_NULL)


    def get_absolute_url(self):
        URL = ':'.join([DRIVER27_NAMESPACE, 'season:race-view'])
        season = self.season
        slug = season.competition.slug
        year = season.year
        return reverse(URL, kwargs = {'competition_slug': slug, 'year': year, 'race_id': self.round})

    def _grandprix_exception(self):
        """ Grand Prix must be related with season competition """
        competition = self.season.competition
        if competition and self.grand_prix and competition not in self.grand_prix.competitions.all():
            return True
        else:
            return False

    def clean(self, *args, **kwargs):
        """ Validate round and grand_prix field """
        errors = {}
        season = getattr(self, 'season', None)
        if season:
            if self._grandprix_exception():
                errors['grand_prix'] = _("%(grand_prix)s is not a/an %(competition)s Grand Prix") \
                                       % {'grand_prix': self.grand_prix, 'competition': self.season.competition}
        if errors:
            raise ValidationError(errors)
        super(Race, self).clean()

    def get_result_seat(self, **kwargs):
        """ Return the first result that match with filter """
        results = self.results.filter(**kwargs)
        return results.first().seat if results.count() else None

    def _abstract_seats(self, exclude=False):
        seat_filter = {'team__competitions__seasons__races': self}
        seat_exclude = {'driver__seats__results__race': self}
        seats = Seat.objects.filter(**seat_filter)
        if exclude:
            seats = seats.exclude(**seat_exclude)

        return seats.filter(
            Q(periods__from_year__lte=self.season.year) | Q(periods__from_year__isnull=True),
            Q(periods__until_year__gte=self.season.year) | Q(periods__until_year__isnull=True)
        ).distinct()

    @property
    def seats(self):
        return self._abstract_seats()

    @property
    def drivers(self):
        return Driver.objects.filter(seats__in=self.seats)

    @property
    def teams(self):
        return Team.objects.filter(seats__in=self.seats).distinct()

    @property
    def no_seats(self):
        return self._abstract_seats(exclude=True)

    @property
    def pole(self):
        """ get_result_seat(pole) """
        return self.get_result_seat(qualifying=1)

    @property
    def winner(self):
        """ get_result_seat(winner) """
        return self.get_result_seat(finish=1)

    @property
    def fastest(self):
        """ get_result_seat(fastest) """
        return self.fastest_car

    @classmethod
    def bulk_copy(cls, races_pk, season_pk):
        races = cls.objects.filter(pk__in=races_pk)
        season = Season.objects.get(pk=season_pk)

        season_rounds = season.rounds
        season_races = season.races.all()

        # As the value of race.round is limited from 1 to season.rounds value,
        # we will find the values available for the races that we are going to incorporate.
        season_rounds_range = list(range(1, season_rounds + 1))
        season_unavailable_rounds = list(season_races.values_list('round', flat=True))
        season_available_rounds = lr_diff(season_rounds_range, season_unavailable_rounds)

        for index, race in enumerate(races):
            # copy the grand_prix and circuit for each original race
            # and set round value between the available values
            cls.objects.create(
                round=season_available_rounds[index],
                season=season,
                grand_prix=race.grand_prix,
                circuit=race.circuit,
            )

    @classmethod
    def check_list_in_season(cls, races_pk, season_pk):
        # To consider that a race is included in a season, the grand prize has to be in it.
        races = cls.objects.filter(pk__in=races_pk)
        races_gp = list(races.values_list('grand_prix_id', flat=True))

        season = Season.objects.get(pk=season_pk)
        season_races = season.races.all()
        season_races_gp = list(season_races.values_list('grand_prix_id', flat=True))

        not_exists = lr_diff(races_gp, season_races_gp)
        both_exists = lr_intr(races_gp, season_races_gp)

        not_exists_races = [race for race in races if race.grand_prix_id in not_exists]
        both_exists_races = [race for race in races if race.grand_prix_id in both_exists]
        can_save = season.rounds >= (season_races.count() + len(not_exists_races))

        return {
            'not_exists': not_exists_races,
            'both_exists': both_exists_races,
            'can_save': can_save,
            'season_info': season
        }

    def __str__(self):
        race_str = '{season}-{round}'.format(season=self.season, round=self.round)
        if self.grand_prix:
            race_str += u'.{grand_prix}'.format(grand_prix=self.grand_prix)
        return race_str

    class Meta:
        unique_together = ('season', 'round')
        ordering = ['season', 'round']
        verbose_name = _('Race')
        verbose_name_plural = _('Races')


@python_2_unicode_compatible
class TeamSeason(TeamStatsModel, SeasonStatsModel):
    """ TeamSeason model, only for validation although it is saved in DB"""
    season = models.ForeignKey('Season', related_name='teams_season', verbose_name=_('season'))
    team = models.ForeignKey('Team', related_name='seasons_team', verbose_name=_('team'))
    sponsor_name = models.CharField(max_length=75, null=True, blank=True, default=None,
                                    verbose_name=_('sponsor name'))

    @property
    def stats_filter_kwargs(self):
        return {'season': self.season, 'team': self.team}

    def get_results(self, limit_races=None, **extra_filter):
        """ Return all results of team in season """
        results_filter = self.stats_filter_kwargs
        return Result.wizard(limit_races=limit_races, **dict(results_filter, **extra_filter))

    def get_points_list(self, limit_races=None, punctuation_config=None, **kwargs):
        points_to_rank = kwargs.pop('points_to_rank', None)
        if punctuation_config is None:
            points_list = self.get_saved_points(limit_races=limit_races, **kwargs)
        else:
            points_list = []
            if points_to_rank:
                results = self.season.get_results_list('team')[self.team]
            else:
                results = self.get_results(limit_races=limit_races, **kwargs)

            for result_tuple in get_tuples_from_results(results=results):
                points = PointsCalculator(punctuation_config).calculator(result_tuple, skip_wildcard=True)
                if points: points_list.append(points)
        return points_list

    def get_points(self, limit_races=None, punctuation_config=None, **kwargs):
        """ Get points. Can be limited. Punctuation config will be overwrite temporarily"""
        points_list = self.get_points_list(limit_races=limit_races, punctuation_config=punctuation_config, **kwargs)

        return sum(points_list)

    def get_name_in_season(self):
        str_team = self.team.name
        if self.sponsor_name:
            str_team = self.sponsor_name
        return str_team

    def _summary_season(self, exclude_position=False):

        summary_season = {
            'season': self.season,
            'competition': self.season.competition,
            'year': self.season.year
        }

        if not exclude_position:
            summary_season['pos'] = self.get_points_position()
        return summary_season

    def get_summary_stats(self, records_list=None, append_points=False, **kwargs):
        """
        Return multiple records (or only one) record in season

        """
        exclude_position = kwargs.pop('exclude_position', False)
        summary_stats = self._summary_season(exclude_position)
        season = summary_stats.get('season')
        seats = season.seats.filter(team=self.team)

        summary_stats.update(
            seats=seats,
            stats=self.team.get_stats_list(records_list=records_list, append_points=append_points, season=season, **kwargs)
        )

        return summary_stats


    def get_points_position(self):

        """
        Return position in season rank

        """
        return self._points_position('team_points_rank', 'team')

    def __str__(self):
        str_team = self.get_name_in_season()
        return '{team} in {season}'.format(team=str_team, season=self.season)

    class Meta:
        unique_together = ('season', 'team')
        verbose_name = _('Team Season')
        verbose_name_plural = _('Teams Season')


@python_2_unicode_compatible
class Result(models.Model):
    """ Result model """
    race = models.ForeignKey(Race, related_name='results', verbose_name=_('race'))
    seat = models.ForeignKey(Seat, related_name='results', verbose_name=_('seat'))
    qualifying = SwapIntegerField(unique_for_fields=['race'], blank=True, null=True, default=None,
                                  verbose_name=_('qualifying'))
    finish = SwapIntegerField(unique_for_fields=['race'], blank=True, null=True, default=None, verbose_name=_('finish'))
    # fastest_lap = ExclusiveBooleanField(on='race', default=False, verbose_name=_('fastest lap'))
    retired = models.BooleanField(default=False, verbose_name=_('retired'))
    wildcard = models.BooleanField(default=False, verbose_name=_('wildcard'))
    comment = models.CharField(max_length=250, blank=True, null=True, default=None, verbose_name=_('comment'))
    points = models.IntegerField(default=0, blank=True, null=True, verbose_name=_('points'))

    @classmethod
    def wizard(cls, seat=None, driver=None, team=None,
               race=None, season=None, competition=None,
               grand_prix=None, circuit=None,
               limit_races=None,
               reverse_order=False,
               **extra_filters):
        key_in_filters = {
            'limit_races': 'race__round__lte',
            'driver': 'seat__driver',
            'team': 'seat__team',
            'season': 'race__season',
            'competition': 'race__season__competition',
            'seat': 'seat',
            'race': 'race',
            'grand_prix': 'race__grand_prix',
            'circuit': 'race__circuit'
        }
        filter_params = {}

        # kwarg is each param with custom filter

        for kwarg in key_in_filters.keys():
            value = locals().get(kwarg)  # get the value of kwarg
            if value:
                key = key_in_filters.get(kwarg)  # the key of filter is the value of kwarg in key_in_filter
                filter_params[key] = value
        filter_params.update(**extra_filters)

        results = cls.objects.filter(**filter_params)

        order_by_args = ('race__season__year', 'race__date', 'race__season__pk', 'race__round') \
            if not reverse_order else ('-race__season__year', '-race__date', '-race__season__pk', '-race__round')

        results = results.order_by(*order_by_args)
        return results

    def _validate_seat(self):
        errors = {'seat': []}
        if not self.race.season.competition.teams.filter(pk=self.seat.team.pk).exists():
            errors['seat'].append('Team not in Competition')
        if Result.wizard(driver=self.seat.driver, race=self.race).exclude(pk=self.pk).exists():
            errors['seat'].append('Exists a result with the same driver in this race (different Seat)')
        if not self.race.season.seats.filter(pk=self.seat.pk).exists():
            errors['seat'].append('{seat} is not valid in {season_year}'.format(seat=self.seat,
                                                                                season_year=self.race.season.year))
        if errors['seat']:
            raise ValidationError(errors)

    def clean(self):
        self._validate_seat()
        super(Result, self).clean()

    def save(self, *args, **kwargs):
        self._validate_seat()
        self.points = self.get_points()
        super(Result, self).save(*args, **kwargs)
        cache.clear()

    @property
    def driver(self):
        return self.seat.driver

    @property
    def team(self):
        return self.seat.team

    @property
    def is_fastest(self):
        try:
            return self.race.fastest_car.pk == self.seat.pk
        except AttributeError:
            return False

    def points_calculator(self, punctuation_config):
        result_tuple = get_tuple_from_result(self)
        calculator = PointsCalculator(punctuation_config).calculator(result_tuple)
        return calculator

    def get_points(self):
        """ Return points based on season scoring """
        punctuation_config = self.race.season.get_punctuation_config()
        return self.points_calculator(punctuation_config)

    def __str__(self):
        string = '{seat} ({race})'.format(seat=self.seat, race=self.race)
        if self.finish:
            string += ' - {finish}º'.format(finish=self.finish)
        else:
            string += ' - ' + _('DNF')
        return string

    class Meta:
        unique_together = ('race', 'seat')
        ordering = ['race__season', 'race__round', 'finish', 'qualifying']
        verbose_name = _('Result')
        verbose_name_plural = _('Results')


class ContenderSeason(AbstractStreakModel, SeasonStatsModel):
    """ ContenderSeason is not a model. Only for validation and ranks"""

    driver = None
    season = None
    teams = None
    contender_teams = None
    teams_verbose = None

    def __init__(self, driver, season):
        self.driver = driver
        self.season = season
        self.seats = Seat.objects.filter(driver__pk=self.driver.pk, results__race__season__pk=self.season.pk)
        self.teams = Team.objects.filter(seats__in=self.seats)
        self.teams_verbose = self.get_teams_verbose()

    def get_team_name_in_season(self, team):
        team_name = team.name
        team_season = TeamSeason.objects.filter(team=team, season=self.season)
        if team_season.exists():
            team_name = '{team}'.format(team=team_season.first().get_name_in_season())
        return team_name

    def get_teams_verbose(self):
        # In case of sponsor_name, it will be show in teams_verbose
        teams_verbose = []
        for team in self.teams:
            team_verbose = self.get_team_name_in_season(team=team)
            teams_verbose.append(team_verbose)
        return ', '.join(teams_verbose)

    def get_results(self, limit_races=None, **extra_filter):
        """ Return results. Can be limited."""
        extra_filter.update(**{'season': self.season})
        return Result.wizard(driver=self.driver, limit_races=limit_races, **extra_filter)

    def get_reverse_results(self, limit_races=None, **extra_filter):
        return self.get_results(limit_races=limit_races, reverse_order=True, **extra_filter)

    def get_stats(self, **filters):
        """ Count 1 by each result """
        return self.get_results(**filters).count()

    def get_positions_in_row(self, limit_races=None, **extra_filter):
        results = self.get_results(limit_races=limit_races, **extra_filter)
        results = get_tuples_from_results(results)
        season_races = list(self.season.past_races.values_list('round', flat=True))
        positions_by_round = {result.round: result.finish for result in results}
        return [positions_by_round[x] for x in season_races if x in positions_by_round]

    def get_saved_points(self, limit_races=None):
        results = self.season.get_results_list()[self.driver]
        return [result.points for result in results if result.points]

    @property
    def is_active(self):
        return True

    def get_points_list(self, limit_races=None, punctuation_config=None, **kwargs):
        points_to_rank = kwargs.pop('points_to_rank', None)
        if punctuation_config is None:
            points_list = self.get_saved_points(limit_races=limit_races)
        else:
            points_list = []
            if points_to_rank:
                results = self.season.get_results_list()[self.driver]
            else:
                results = self.get_results(limit_races=limit_races, **kwargs)
            for result_tuple in get_tuples_from_results(results=results):
                points = PointsCalculator(punctuation_config).calculator(result_tuple)
                if points: points_list.append(points)
        return points_list

    def get_points(self, limit_races=None, punctuation_config=None, **kwargs):
        """ Get points. Can be limited. Punctuation config will be overwrite temporarily"""

        points_list = self.get_points_list(limit_races=limit_races, punctuation_config=punctuation_config, **kwargs)
        points_list = sorted(points_list, reverse=True)
        season_rounds = self.season.rounds
        if season_rounds:
            points_list = points_list[:season_rounds]
        return sum(points_list)

    def _summary_season(self, exclude_position=False):
        summary_season = {
            'season': self.season,
            'competition': self.season.competition,
            'year': self.season.year,
            'teams': self.teams_verbose
        }

        if not exclude_position:
            summary_season['pos'] = self.get_points_position()

        return summary_season



    def get_summary_stats(self, records_list=None, append_points=False, **kwargs):
        """
        Return multiple records (or only one) record in season

        """
        exclude_position = kwargs.pop('exclude_position', False)
        summary_stats = self._summary_season(exclude_position)
        summary_stats.update(
            stats=self.driver.get_stats_list(records_list=records_list, append_points=append_points,
                                                   season=self.season, **kwargs)
        )
        return summary_stats

    def get_points_position(self):

        """
        Return position in season rank

        """
        return self._points_position('points_rank', 'driver')


class RankModel(AbstractRankModel):
    @property
    def stats_filter_kwargs(self):
        return {}

    def get_stats_cls(self, driver):
        return driver

    def get_team_stats_cls(self, team):
        return team

    @property
    def drivers(self):
        return Driver.objects.all()

    @property
    def teams(self):
        return Team.objects.all()

    def __str__(self):
        return _('Global rank')

    class Meta:
        abstract = True



from django.core.management.base import BaseCommand, CommandError
from driver27.models import Driver, Competition, Season
from driver27.records import get_record_config


class Command(BaseCommand):
    help = 'Test multirecord'

    def get_multirecord(self, driver, multirecords, **kwargs):
        multiple_records = {}
        for multirecord in multirecords:
            record_config = get_record_config(multirecord).get('filter')
            record_stat = driver.get_stats(**dict(kwargs, **record_config))
            multiple_records[multirecord] = record_stat
        return multiple_records

    def handle(self, *args, **options):
        driver = Driver.objects.get(pk=1)
        multirecords = ['PODIUM', 'POLE', 'WIN', 'FASTEST']

        competition = Competition.objects.get(pk=1)
        season_1 = Season.objects.get(pk=1)
        season_3 = Season.objects.get(pk=3)

        print(self.get_multirecord(driver, multirecords=multirecords, competition=competition))
        print(self.get_multirecord(driver, multirecords=multirecords, season=season_1))
        print(self.get_multirecord(driver, multirecords=multirecords, season=season_3))




from django.core.management.base import BaseCommand, CommandError
from driver27.models import Result
import csv


class Command(BaseCommand):
    help = 'Update results via csv'

    @staticmethod
    def _create_result(row):
        invalid_keywords = []
        result_fields = [f.column for f in Result._meta.get_fields() if hasattr(f, 'column')]
        for x in row:
            if x in ['qualifying', 'finish', 'points', 'race_id', 'seat_id']:
                if row[x] == '':
                    row[x] = None
                else:
                    row[x] = int(row[x])
            elif x in ['wildcard', 'fastest_lap', 'retired']:
                row[x] = bool(row[x])
            elif x not in result_fields:
                invalid_keywords.append(x)

        for invalid_keyword in invalid_keywords:
            del row[invalid_keyword]
        return Result.objects.create(**row)

    def add_arguments(self, parser):
        parser.add_argument('csv',)

    def handle(self, *args, **options):
        with open(options['csv'], 'rb') as csvfile:
            reader = csv.DictReader(csvfile, delimiter=',')
            for row in reader:
                self._create_result(row)



from .config import get_config


def get_record_config(record_code=None):
    config = get_config('RECORDS', key=record_code)
    return config


def get_record_label_dict(doubles=False):
    def _get_double_records():
        return [(index, record_dict.get('label')) for index, record_dict in config.items()
                if record_dict.get('doubles')]

    def _get_records():
        return [(index, record_dict.get('label')) for index, record_dict in config.items()]

    config = get_record_config()
    if config:
        return _get_double_records() if doubles else _get_records()
    return None



from django.conf.urls import url, include

from driver27 import views


class Driver27ShortcutUrls(list):
    def __init__(self, base_name):
        self.base_name = base_name
        super(Driver27ShortcutUrls, self).__init__()

    def bulk_append_urls(self, urls_list):
        for url_entry in urls_list:
            self.append_url(*url_entry)

    def append_url(self, pattern, view_name, name, kwargs=None):
        if kwargs is None: kwargs = {}
        url_name = '-'.join([self.base_name, name]) if name else self.base_name
        self.append(
            url(r'^' + pattern, getattr(views, view_name), name=url_name, kwargs=kwargs)
        )


def dr27_record_urls_by_type(stat_type):
    return {
        'driver': [
            ('streak/actives/$', stat_type + '_streak_view', 'active-streak', {'only_actives': True}),
            ('streak/top-actives/$', stat_type + '_streak_view', 'active-top-streak', {'only_actives': True,
                                                                                       'max_streak': True})
        ],
        'team': [
            ('races/$', 'team_record_view', 'record-races', {'rank_type': 'RACES'}),
            ('doubles-in-race/$', 'team_record_view', 'record-doubles', {'rank_type': 'DOUBLES'}),
        ]
    }.get(stat_type)


def dr27_stats_urls(stat_type):
    url_pattern = Driver27ShortcutUrls(stat_type)
    url_pattern.bulk_append_urls(
        [
            ('$', '_rank_view', '', {'rank_model': stat_type}),
            ('seasons/$', '_rank_view', 'seasons-rank', {'rank_model': stat_type, 'by_season': True}),
            ('olympic/$', stat_type + '_olympic_view', 'olympic'),
            ('record/$', stat_type + '_record_view', 'record-index'),
        ]
    )

    if stat_type == 'driver':
        url_pattern.append_url('comeback/$', 'driver_comeback_view', 'comeback')

    return url_pattern



from django.shortcuts import render, get_object_or_404


def competition_request(f):
    def wrap(request, *args, **kwargs):
        if request.is_ajax():
            slug = request.GET.get('competition_slug', None)
            year = request.GET.get('year', None)
        else:
            slug = kwargs.pop('competition_slug', None)
            year = kwargs.pop('year', None)

        from .models import Season, Competition, RankModel
        season = competition = None
        if slug:
            if year:
                season_or_competition = get_object_or_404(Season, competition__slug=slug, year=year)
                season = season_or_competition
                competition = season.competition
            else:
                season_or_competition = get_object_or_404(Competition, slug=slug)
                competition = season_or_competition
        else:
            season_or_competition = RankModel()

        context = kwargs.setdefault('context', {})
        context.update(season=season, competition=competition, season_or_competition=season_or_competition)
        kwargs['context'] = context
        return f(request, *args, **kwargs)

    wrap.__doc__ = f.__doc__
    wrap.__name__ = f.__name__
    return wrap



from django.conf.urls import url, include

from driver27 import views

urlpatterns = [
    url(r'^$', views.competition_view, name='view'),
    url(r'^driver/', include('driver27.urls.driver')),
    url(r'^team/', include('driver27.urls.team')),
]



# -*- coding: utf-8 -*-
from django import template
register = template.Library()

@register.filter
def print_pos(pos):
    if pos:
        str_pos = u'{pos}º'.format(pos=pos)
        if pos == 1:
            str_pos = '<strong>' + str_pos + '</strong>'
    else:
        str_pos = u''
    return str_pos



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-24 20:58
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_countries.fields
import swapfield.fields


def populate_driver_in_seats(apps, schema_editor):
    Seat = apps.get_model("driver27", "Seat")
    for seat in Seat.objects.all():
        driver = seat.contender.driver
        seat.driver = driver
        seat.save()

def seats_contender_null(apps, schema_editor):
    Seat = apps.get_model("driver27", "Seat")
    for seat in Seat.objects.all():
        seat.contender = None
        seat.save()

def populate_competition_team(apps, schema_editor):
    Team = apps.get_model("driver27", "Team")
    CompetitionTeam = apps.get_model("driver27", "CompetitionTeam")
    for team in Team.objects.all():
        competitions = team.competitions.all()
        for competition in competitions:
            CompetitionTeam.objects.create(team=team, competition=competition)

def populate_fastest_lap(apps, schema_editor):
    Race = apps.get_model("driver27", "Race")
    for race in Race.objects.all():
        fastest = race.results.filter(fastest_lap=True).first()
        if fastest:
            race.fastest_car = fastest.seat
            race.save()

class Migration(migrations.Migration):

    replaces = [(b'driver27', '0001_initial'), (b'driver27', '0002_auto_20161103_0846'), (b'driver27', '0003_auto_20170204_2308'), (b'driver27', '0004_auto_20170204_2311'), (b'driver27', '0005_result_points'), (b'driver27', '0006_auto_20170529_2209'), (b'driver27', '0007_auto_20170529_2211'), (b'driver27', '0008_auto_20170529_2220'), (b'driver27', '0009_populate_driver_in_seats'), (b'driver27', '0010_seats_contender_null'), (b'driver27', '0011_remove_contender_teams'), (b'driver27', '0012_auto_20170529_2232'), (b'driver27', '0013_remove_seat_contender'), (b'driver27', '0014_auto_20170529_2320'), (b'driver27', '0015_auto_20170602_1447'), (b'driver27', '0016_driver_teams'), (b'driver27', '0017_auto_20170602_1526'), (b'driver27', '0018_populate_competition_team'), (b'driver27', '0019_remove_team_competitions'), (b'driver27', '0020_team_competitions'), (b'driver27', '0021_auto_20170602_1547'), (b'driver27', '0022_auto_20170602_1552'), (b'driver27', '0023_auto_20170602_1552'), (b'driver27', '0024_auto_20170602_1737'), (b'driver27', '0025_auto_20171022_2012'), (b'driver27', '0026_populate_fastest_race'), (b'driver27', '0027_auto_20171022_2021')]

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Circuit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='circuit')),
                ('city', models.CharField(blank=True, max_length=100, null=True)),
                ('country', django_countries.fields.CountryField(max_length=2)),
                ('opened_in', models.IntegerField()),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Competition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='competition')),
                ('full_name', models.CharField(max_length=100, unique=True)),
                ('country', django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True)),
                ('slug', models.SlugField(blank=True, default=None, null=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Contender',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contenders', to='driver27.Competition')),
            ],
            options={
                'ordering': ['competition__name', 'driver__last_name', 'driver__first_name'],
            },
        ),
        migrations.CreateModel(
            name='Driver',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_name', models.CharField(max_length=50)),
                ('first_name', models.CharField(max_length=25)),
                ('year_of_birth', models.IntegerField()),
                ('country', django_countries.fields.CountryField(max_length=2)),
                ('competitions', models.ManyToManyField(related_name='drivers', through='driver27.Contender', to=b'driver27.Competition')),
            ],
            options={
                'ordering': ['last_name', 'first_name'],
            },
        ),
        migrations.CreateModel(
            name='GrandPrix',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='grand prix')),
                ('country', django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True)),
                ('first_held', models.IntegerField(blank=True, null=True)),
                ('competitions', models.ManyToManyField(default=None, related_name='grands_prix', to=b'driver27.Competition')),
                ('default_circuit', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='default_to_grands_prix', to='driver27.Circuit')),
            ],
            options={
                'ordering': ['name'],
                'verbose_name_plural': 'grands prix',
            },
        ),
        migrations.CreateModel(
            name='Race',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('round', models.IntegerField()),
                ('date', models.DateField(blank=True, default=None, null=True)),
                ('alter_punctuation', models.CharField(blank=True, choices=[('double', 'Double'), ('half', 'Half')], default=None, max_length=6, null=True)),
                ('circuit', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Circuit')),
                ('grand_prix', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.GrandPrix')),
            ],
            options={
                'ordering': ['season', 'round'],
            },
        ),
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('qualifying', models.IntegerField(blank=True, default=None, null=True)),
                ('finish', models.IntegerField(blank=True, default=None, null=True)),
                ('fastest_lap', models.BooleanField(default=False)),
                ('retired', models.BooleanField(default=False)),
                ('wildcard', models.BooleanField(default=False)),
                ('comment', models.CharField(blank=True, default=None, max_length=250, null=True)),
                ('race', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Race')),
            ],
            options={
                'ordering': ['race__season', 'race__round', 'finish', 'qualifying'],
            },
        ),
        migrations.CreateModel(
            name='Season',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('year', models.IntegerField()),
                ('rounds', models.IntegerField(blank=True, default=None, null=True)),
                ('punctuation', models.CharField(default=None, max_length=20, null=True)),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons', to='driver27.Competition')),
            ],
            options={
                'ordering': ['competition__name', 'year'],
            },
        ),
        migrations.CreateModel(
            name='Seat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('current', models.BooleanField(default=False)),
                ('contender', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Contender')),
                ('seasons', models.ManyToManyField(blank=True, default=None, related_name='seats', to=b'driver27.Season')),
            ],
            options={
                'ordering': ['contender__driver__last_name', 'team'],
            },
        ),
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=75, unique=True, verbose_name='team')),
                ('full_name', models.CharField(max_length=200, unique=True)),
                ('country', django_countries.fields.CountryField(max_length=2)),
                ('competitions', models.ManyToManyField(related_name='teams', to=b'driver27.Competition')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='TeamSeason',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sponsor_name', models.CharField(blank=True, default=None, max_length=75, null=True)),
                ('season', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='teams_season', to='driver27.Season')),
                ('team', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons_team', to='driver27.Team')),
            ],
            options={
                'verbose_name': 'Team Season',
                'verbose_name_plural': 'Teams Season',
            },
        ),
        migrations.AddField(
            model_name='seat',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Team'),
        ),
        migrations.AddField(
            model_name='season',
            name='teams',
            field=models.ManyToManyField(related_name='seasons', through='driver27.TeamSeason', to=b'driver27.Team'),
        ),
        migrations.AddField(
            model_name='result',
            name='seat',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Seat'),
        ),
        migrations.AddField(
            model_name='race',
            name='season',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Season'),
        ),
        migrations.AddField(
            model_name='contender',
            name='driver',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='career', to='driver27.Driver'),
        ),
        migrations.AddField(
            model_name='contender',
            name='teams',
            field=models.ManyToManyField(related_name='contenders', through='driver27.Seat', to=b'driver27.Team'),
        ),
        migrations.AlterUniqueTogether(
            name='teamseason',
            unique_together=set([('season', 'team')]),
        ),
        migrations.AlterField(
            model_name='seat',
            name='current',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterUniqueTogether(
            name='seat',
            unique_together=set([('team', 'contender')]),
        ),
        migrations.AlterUniqueTogether(
            name='season',
            unique_together=set([('year', 'competition')]),
        ),
        migrations.AlterField(
            model_name='result',
            name='fastest_lap',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterUniqueTogether(
            name='result',
            unique_together=set([('race', 'seat'), ('race', 'qualifying'), ('race', 'finish')]),
        ),
        migrations.AlterUniqueTogether(
            name='race',
            unique_together=set([('season', 'round')]),
        ),
        migrations.AlterUniqueTogether(
            name='driver',
            unique_together=set([('last_name', 'first_name')]),
        ),
        migrations.AlterUniqueTogether(
            name='contender',
            unique_together=set([('driver', 'competition')]),
        ),
        migrations.CreateModel(
            name='SeatsSeason',
            fields=[
            ],
            options={
                'verbose_name': 'Seats by season',
                'proxy': True,
                'verbose_name_plural': 'Seats by season',
            },
            bases=('driver27.season',),
        ),
        migrations.AlterModelOptions(
            name='circuit',
            options={'ordering': ['name'], 'verbose_name': 'Circuit', 'verbose_name_plural': 'Circuits'},
        ),
        migrations.AlterModelOptions(
            name='competition',
            options={'ordering': ['name'], 'verbose_name': 'Competition', 'verbose_name_plural': 'Competitions'},
        ),
        migrations.AlterModelOptions(
            name='contender',
            options={'ordering': ['competition__name', 'driver__last_name', 'driver__first_name'], 'verbose_name': 'Contender', 'verbose_name_plural': 'Contenders'},
        ),
        migrations.AlterModelOptions(
            name='driver',
            options={'ordering': ['last_name', 'first_name'], 'verbose_name': 'Driver', 'verbose_name_plural': 'Drivers'},
        ),
        migrations.AlterModelOptions(
            name='grandprix',
            options={'ordering': ['name'], 'verbose_name': 'Grand Prix', 'verbose_name_plural': 'Grands Prix'},
        ),
        migrations.AlterModelOptions(
            name='race',
            options={'ordering': ['season', 'round'], 'verbose_name': 'Race', 'verbose_name_plural': 'Races'},
        ),
        migrations.AlterModelOptions(
            name='result',
            options={'ordering': ['race__season', 'race__round', 'finish', 'qualifying'], 'verbose_name': 'Result', 'verbose_name_plural': 'Results'},
        ),
        migrations.AlterModelOptions(
            name='season',
            options={'ordering': ['competition__name', 'year'], 'verbose_name': 'Season', 'verbose_name_plural': 'Seasons'},
        ),
        migrations.AlterModelOptions(
            name='seat',
            options={'ordering': ['current', 'contender__driver__last_name', 'team'], 'verbose_name': 'Seat', 'verbose_name_plural': 'Seats'},
        ),
        migrations.AlterModelOptions(
            name='team',
            options={'ordering': ['name'], 'verbose_name': 'Team', 'verbose_name_plural': 'Teams'},
        ),
        migrations.AlterField(
            model_name='circuit',
            name='city',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='city'),
        ),
        migrations.AlterField(
            model_name='circuit',
            name='country',
            field=django_countries.fields.CountryField(max_length=2, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='circuit',
            name='opened_in',
            field=models.IntegerField(verbose_name='opened in'),
        ),
        migrations.AlterField(
            model_name='competition',
            name='country',
            field=django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='competition',
            name='full_name',
            field=models.CharField(max_length=100, unique=True, verbose_name='full name'),
        ),
        migrations.AlterField(
            model_name='contender',
            name='competition',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contenders', to='driver27.Competition', verbose_name='competition'),
        ),
        migrations.AlterField(
            model_name='contender',
            name='driver',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='career', to='driver27.Driver', verbose_name='driver'),
        ),
        migrations.AlterField(
            model_name='contender',
            name='teams',
            field=models.ManyToManyField(related_name='contenders', through='driver27.Seat', to=b'driver27.Team', verbose_name='teams'),
        ),
        migrations.AlterField(
            model_name='driver',
            name='competitions',
            field=models.ManyToManyField(related_name='drivers', through='driver27.Contender', to=b'driver27.Competition', verbose_name='competitions'),
        ),
        migrations.AlterField(
            model_name='driver',
            name='country',
            field=django_countries.fields.CountryField(max_length=2, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='driver',
            name='first_name',
            field=models.CharField(max_length=25, verbose_name='first name'),
        ),
        migrations.AlterField(
            model_name='driver',
            name='last_name',
            field=models.CharField(max_length=50, verbose_name='last name'),
        ),
        migrations.AlterField(
            model_name='driver',
            name='year_of_birth',
            field=models.IntegerField(verbose_name='year of birth'),
        ),
        migrations.AlterField(
            model_name='grandprix',
            name='competitions',
            field=models.ManyToManyField(default=None, related_name='grands_prix', to=b'driver27.Competition', verbose_name='competitions'),
        ),
        migrations.AlterField(
            model_name='grandprix',
            name='country',
            field=django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='grandprix',
            name='default_circuit',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='default_to_grands_prix', to='driver27.Circuit', verbose_name='default circuit'),
        ),
        migrations.AlterField(
            model_name='grandprix',
            name='first_held',
            field=models.IntegerField(blank=True, null=True, verbose_name='first held'),
        ),
        migrations.AlterField(
            model_name='race',
            name='alter_punctuation',
            field=models.CharField(blank=True, choices=[('double', 'Double'), ('half', 'Half')], default=None, max_length=6, null=True, verbose_name='alter punctuation'),
        ),
        migrations.AlterField(
            model_name='race',
            name='circuit',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Circuit', verbose_name='circuit'),
        ),
        migrations.AlterField(
            model_name='race',
            name='date',
            field=models.DateField(blank=True, default=None, null=True, verbose_name='date'),
        ),
        migrations.AlterField(
            model_name='race',
            name='grand_prix',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.GrandPrix', verbose_name='grand prix'),
        ),
        migrations.AlterField(
            model_name='race',
            name='round',
            field=models.IntegerField(verbose_name='round'),
        ),
        migrations.AlterField(
            model_name='race',
            name='season',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Season', verbose_name='season'),
        ),
        migrations.AlterField(
            model_name='result',
            name='comment',
            field=models.CharField(blank=True, default=None, max_length=250, null=True, verbose_name='comment'),
        ),
        migrations.AlterField(
            model_name='result',
            name='fastest_lap',
            field=models.BooleanField(default=False, verbose_name='fastest lap'),
        ),
        migrations.AlterField(
            model_name='result',
            name='finish',
            field=models.IntegerField(blank=True, default=None, null=True, verbose_name='finish'),
        ),
        migrations.AlterField(
            model_name='result',
            name='qualifying',
            field=models.IntegerField(blank=True, default=None, null=True, verbose_name='qualifying'),
        ),
        migrations.AlterField(
            model_name='result',
            name='race',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Race', verbose_name='race'),
        ),
        migrations.AlterField(
            model_name='result',
            name='retired',
            field=models.BooleanField(default=False, verbose_name='retired'),
        ),
        migrations.AlterField(
            model_name='result',
            name='seat',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Seat', verbose_name='seat'),
        ),
        migrations.AlterField(
            model_name='result',
            name='wildcard',
            field=models.BooleanField(default=False, verbose_name='wildcard'),
        ),
        migrations.AlterField(
            model_name='season',
            name='competition',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons', to='driver27.Competition', verbose_name='competition'),
        ),
        migrations.AlterField(
            model_name='season',
            name='punctuation',
            field=models.CharField(default=None, max_length=20, null=True, verbose_name='punctuation'),
        ),
        migrations.AlterField(
            model_name='season',
            name='rounds',
            field=models.IntegerField(blank=True, default=None, null=True, verbose_name='rounds'),
        ),
        migrations.RemoveField(
            model_name='season',
            name='teams',
        ),
        migrations.AlterField(
            model_name='season',
            name='year',
            field=models.IntegerField(verbose_name='year'),
        ),
        migrations.AlterField(
            model_name='seat',
            name='contender',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Contender', verbose_name='contender'),
        ),
        migrations.AlterField(
            model_name='seat',
            name='current',
            field=models.BooleanField(default=False, verbose_name='current team'),
        ),
        migrations.AlterField(
            model_name='seat',
            name='seasons',
            field=models.ManyToManyField(blank=True, default=None, related_name='seats', to=b'driver27.Season', verbose_name='seasons'),
        ),
        migrations.AlterField(
            model_name='seat',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Team', verbose_name='team'),
        ),
        migrations.AlterField(
            model_name='team',
            name='competitions',
            field=models.ManyToManyField(related_name='teams', to=b'driver27.Competition', verbose_name='competitions'),
        ),
        migrations.AlterField(
            model_name='team',
            name='country',
            field=django_countries.fields.CountryField(max_length=2, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='team',
            name='full_name',
            field=models.CharField(max_length=200, unique=True, verbose_name='full name'),
        ),
        migrations.AlterField(
            model_name='teamseason',
            name='season',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='teams_season', to='driver27.Season', verbose_name='season'),
        ),
        migrations.AlterField(
            model_name='teamseason',
            name='sponsor_name',
            field=models.CharField(blank=True, default=None, max_length=75, null=True, verbose_name='sponsor name'),
        ),
        migrations.AlterField(
            model_name='teamseason',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons_team', to='driver27.Team', verbose_name='team'),
        ),
        migrations.AlterField(
            model_name='result',
            name='finish',
            field=swapfield.fields.SwapIntegerField(blank=True, default=None, null=True, unique_for_fields=['race'], verbose_name='finish'),
        ),
        migrations.AlterField(
            model_name='result',
            name='qualifying',
            field=swapfield.fields.SwapIntegerField(blank=True, default=None, null=True, unique_for_fields=['race'], verbose_name='qualifying'),
        ),
        migrations.AddField(
            model_name='result',
            name='points',
            field=models.IntegerField(blank=True, default=0, null=True, verbose_name='points'),
        ),
        migrations.AlterUniqueTogether(
            name='result',
            unique_together=set([('race', 'seat')]),
        ),
        migrations.AlterModelOptions(
            name='seat',
            options={'ordering': ['contender__driver__last_name', 'team'], 'verbose_name': 'Seat', 'verbose_name_plural': 'Seats'},
        ),
        migrations.RemoveField(
            model_name='seat',
            name='current',
        ),
        migrations.AddField(
            model_name='seat',
            name='driver',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Driver', verbose_name='driver'),
        ),
        migrations.AlterField(
            model_name='seat',
            name='contender',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Contender', verbose_name='contender'),
        ),
        migrations.RemoveField(
            model_name='seat',
            name='seasons',
        ),
        migrations.RunPython(
            code=populate_driver_in_seats,
        ),
        migrations.RunPython(
            code=seats_contender_null,
        ),
        migrations.RemoveField(
            model_name='contender',
            name='teams',
        ),
        migrations.AlterModelOptions(
            name='seat',
            options={'ordering': ['driver__last_name', 'team'], 'verbose_name': 'Seat', 'verbose_name_plural': 'Seats'},
        ),
        migrations.RemoveField(
            model_name='seat',
            name='contender',
        ),
        migrations.AlterUniqueTogether(
            name='seat',
            unique_together=set([('team', 'driver')]),
        ),
        migrations.AlterUniqueTogether(
            name='contender',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='contender',
            name='competition',
        ),
        migrations.RemoveField(
            model_name='contender',
            name='driver',
        ),
        migrations.RemoveField(
            model_name='driver',
            name='competitions',
        ),
        migrations.DeleteModel(
            name='Contender',
        ),
        migrations.AlterField(
            model_name='driver',
            name='country',
            field=django_countries.fields.CountryField(blank=True, max_length=2, null=True, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='driver',
            name='year_of_birth',
            field=models.IntegerField(blank=True, null=True, verbose_name='year of birth'),
        ),
        migrations.AddField(
            model_name='driver',
            name='teams',
            field=models.ManyToManyField(related_name='drivers', through='driver27.Seat', to=b'driver27.Team', verbose_name='teams'),
        ),
        migrations.CreateModel(
            name='CompetitionTeam',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_year', models.IntegerField(blank=True, default=None, null=True)),
                ('until_year', models.IntegerField(blank=True, default=None, null=True)),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='driver27.Competition', verbose_name='competition')),
                ('team', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='driver27.Team', verbose_name='team')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='competitionteam',
            unique_together=set([('team', 'competition', 'from_year', 'until_year')]),
        ),
        migrations.RunPython(
            code=populate_competition_team,
        ),
        migrations.RemoveField(
            model_name='team',
            name='competitions',
        ),
        migrations.AddField(
            model_name='team',
            name='competitions',
            field=models.ManyToManyField(related_name='teams', through='driver27.CompetitionTeam', to=b'driver27.Competition', verbose_name='competitions'),
        ),
        migrations.AlterUniqueTogether(
            name='seat',
            unique_together=set([('team', 'driver')]),
        ),
        migrations.CreateModel(
            name='SeatPeriod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_year', models.IntegerField(blank=True, default=None, null=True)),
                ('until_year', models.IntegerField(blank=True, default=None, null=True)),
                ('seat', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='driver27.Seat')),
            ],
        ),
        migrations.AlterField(
            model_name='team',
            name='country',
            field=django_countries.fields.CountryField(blank=True, max_length=2, null=True, verbose_name='country'),
        ),
        migrations.AddField(
            model_name='race',
            name='fastest_car',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='fastest_in', to='driver27.Seat'),
        ),
        migrations.RunPython(
            code=populate_fastest_lap,
        ),
        migrations.RemoveField(
            model_name='result',
            name='fastest_lap',
        ),
    ]



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from django.db import transaction, IntegrityError
from .common import CommonResultTestCase
from ..models import TeamSeason, Seat, CompetitionTeam


class SeasonTestCase(TestCase, CommonResultTestCase):
    def test_season_unicode(self):
        season = self.get_test_season()
        year = season.year
        expected_season = '{competition}/{year}'.format(competition=season.competition, year=year)
        self.assertEquals(str(season), expected_season)

    def test_season_scoring(self):
        season = self.get_test_season()
        self.assertIsInstance(season.get_punctuation_config(), dict)
        season.punctuation = 'BLABLABLA'
        self.assertIsNone(season.save())
        self.assertIsNone(season.get_punctuation_config())

    def test_season_has_champion(self):
        season = self.get_test_season()
        self.assertFalse(season.has_champion())

    def test_season_pending_races(self):
        season = self.get_test_season()
        self.assertIsInstance(season.num_pending_races, int)

    def test_season_pending_points(self):
        season = self.get_test_season()
        self.assertIsInstance(season.pending_points(), (int, float))

    def test_season_contenders(self):
        seat_a = self.get_test_seat()
        seat_b = self.get_test_seat_teammate(seat_a)
        team = seat_a.team
        competition = self.get_test_competition()
        self.assertTrue(CompetitionTeam.objects.create(competition=competition, team=team))
        # season = self.get_test_season(competition)
        season = self.get_test_season(competition=competition)
        race = self.get_test_race(season=season, round=1)
        result_a = self.get_test_result(seat=seat_a, race=race, qualifying=1, finish=2)
        result_b = self.get_test_result(seat=seat_b, race=race, qualifying=4, finish=7)
        self.assertEquals(season.drivers.count(), 2)



import os
from .settings import *

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'local.sqlite3'),
    }
}

PYTEST_SETTING = True
ALLOWED_HOSTS.append('testserver')


from django.conf.urls import url
from django.shortcuts import render
from django.utils.translation import ugettext as _
from .common import CommonTabbedModelAdmin
from django.shortcuts import redirect
from .forms import *
from .inlines import *
from ..models import Driver, Competition, Circuit, Season, Result, CompetitionTeam, SeatPeriod
from .. import lr_diff, lr_intr

from django.contrib.admin import SimpleListFilter

from django.db.models import Q


class DriverAdmin(RelatedCompetitionAdmin, CommonTabbedModelAdmin):
    list_display = ('__str__', 'country',)
    list_filter = ('country', 'year_of_birth', 'teams',)
    tab_overview = (
        (None, {
            'fields': ('last_name', 'first_name', 'year_of_birth', 'country')
        }),
    )
    tabs = [
        ('Overview', tab_overview),
    ]


class CompetitionTeamInline(CompetitionFilterInline):
    model = CompetitionTeam


class TeamAdmin(RelatedCompetitionAdmin, CommonTabbedModelAdmin):
    model = Team
    list_display = ('__str__', 'country', 'print_competitions')
    list_filter = ('competitions',)
    tab_overview = (
        (None, {
            'fields': ('name', 'full_name', 'country')
        }),
    )
    tab_competitions = (
        CompetitionTeamInline,
    )
    tab_seats = (
        SeatInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Competitions', tab_competitions),
        ('Seats', tab_seats),
    ]


class CompetitionSeasonAdmin(admin.TabularInline):
    model = Season


class CompetitionAdmin(CommonTabbedModelAdmin):
    model = Competition
    list_display = ('name', 'full_name')
    tab_overview = (
        (None, {
            'fields': ('name', 'full_name', 'country', 'slug')
        }),
    )
    # tab_team = (
    #     CompetitionTeamInline,
    # )
    tab_seasons = (
        CompetitionSeasonAdmin,
    )
    tabs = [
        ('Overview', tab_overview),
        # ('Team', tab_team),
        ('Seasons', tab_seasons)
    ]


class CircuitAdmin(admin.ModelAdmin):
    list_display = ('name', 'city', 'country', 'opened_in')


class GrandPrixAdmin(RelatedCompetitionAdmin, admin.ModelAdmin):
    list_display = ('name', 'country', 'default_circuit', 'print_competitions')
    list_filter = ('competitions',)


class SeasonAdmin(CommonTabbedModelAdmin):
    form = SeasonAdminForm
    tab_overview = (
        (None, {
            'fields': ('year', 'competition', 'rounds', 'punctuation')
        }),
    )
    tab_races = (
        RaceInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Races', tab_races),
    ]
    readonly_fields = ('print_copy_season',)
    list_display = ('__str__', 'print_copy_season', 'print_copy_races',)
    list_filter = ('competition',)

    def get_season_copy(self, copy_id):
        seasons = Season.objects.filter(pk=copy_id)
        copy_dict = {}
        if seasons.count():
            season = seasons.first()
            copy_dict = {
                # 'year': season.year+1,
                'competition': season.competition,
                'rounds': season.rounds,
                'punctuation': season.punctuation
            }
        return copy_dict

    def get_copy_item(self, request, pk, item_cls, items_plural, *args, **kwargs):
        season = Season.objects.get(pk=pk)
        if request.POST.get('_confirm'):
            post_items_pk = request.POST.getlist('items', [])
            post_season_destiny = request.POST.get('season_destiny', None)
            item_cls.bulk_copy(post_items_pk, post_season_destiny)
            return redirect('admin:driver27_season_change', post_season_destiny)

        elif request.POST.get('_selector'):

            post_season_destiny = request.POST.get('season_destiny', None)
            context = {
                'season': season,
                'opts': self.model._meta,
                'app_label': self.model._meta.app_label,
                'change': True,
                'items_plural': items_plural,
                'title': 'Copy {items_plural} from season {season_slug} > Step 2'.format(items_plural=items_plural,
                                                                                         season_slug=season),
                'step': 2
            }
            if post_season_destiny:
                post_items_pk = request.POST.getlist('items', [])
                check_list_in_season = item_cls.check_list_in_season(post_items_pk, post_season_destiny)

                context.update(
                    {
                        'season_destiny': post_season_destiny,
                        'not_exists': check_list_in_season.get('not_exists'),
                        'both_exists': check_list_in_season.get('both_exists'),
                        'conditional_exists': check_list_in_season.get('conditional_exists', None),
                        'can_save': check_list_in_season.get('can_save'),
                        'season_destiny_info': check_list_in_season.get('season_info')
                    }
                )
            else:
                context['no_destiny'] = True
        else:
            items = getattr(season, items_plural).all()
            available_seasons = Season.objects.filter(competition=season.competition).exclude(pk=pk)

            context = {
                'season': season,
                'items': items,
                'available_seasons': available_seasons,
                'opts': self.model._meta,
                'app_label': self.model._meta.app_label,
                'change': True,
                'items_plural': items_plural,
                'title': 'Copy {items_plural} from season {season_slug}'.format(items_plural=items_plural,
                                                                                season_slug=season),
                'step': 1
            }

        return render(request, 'driver27/admin/copy/copy_items.html', context)

    def get_copy_races(self, request, pk, *args, **kwargs):
        return self.get_copy_item(request, pk, Race, 'races', *args, **kwargs)

    def get_duplicate_season(self, request, pk, *args, **kwargs):
        season = Season.objects.get(pk=pk)

        context = {
            'season': season,
            'opts': self.model._meta,
            'app_label': self.model._meta.app_label,
            'change': True,
            'title': 'Duplicating season {season}'.format(season=season)

        }
        if request.POST.get('_selector'):
            season_year = request.POST.get('year')
            if Season.objects.filter(year=season_year).exists():
                error_message = _("Season with {competition: %(competition)s, year: %(year)s} already exists") % \
                                {'competition': season.competition, 'year': season_year}
                context['error_message'] = error_message
            else:
                new_season, created = Season.objects.get_or_create(competition=season.competition,
                                                          year=season_year,
                                                          rounds=season.rounds,
                                                          punctuation=season.punctuation)

                return redirect('admin:driver27_season_change', new_season.pk)


        return render(request, 'driver27/admin/copy/copy_season.html', context)

    def print_copy_season(self, obj):
        if obj.pk:
            copy_link = reverse("admin:dr27-copy-season", kwargs={'pk': obj.pk})
            return "<a href='{link}'>{copy_text}</a>".format(link=copy_link,
                                                                    copy_text=_('Duplicate season'))
        else:
            return ''

    print_copy_season.short_description = _('Create new season from this')
    print_copy_season.allow_tags = True

    def print_copy_link(self, obj, reverse_link, copy_text):
        if obj.pk:
            copy_link = reverse(reverse_link, kwargs={'pk': obj.pk})
            return "<a href='{link}'>{copy_text}</a>".format(link=copy_link, copy_text=copy_text)
        else:
            return ''

    def print_copy_races(self, obj):
        return self.print_copy_link(obj, "admin:dr27-copy-races", _('copy races'))

    print_copy_races.short_description = _('copy races')
    print_copy_races.allow_tags = True

    def get_urls(self, *args, **kwargs):
        urls = super(SeasonAdmin, self).get_urls(*args, **kwargs)
        new_urls = [
                       url(r'^(?P<pk>[\d]+)/get_copy_races/$', self.admin_site.admin_view(self.get_copy_races),
                           name='dr27-copy-races'),
                       url(r'^(?P<pk>[\d]+)/duplicate/$', self.admin_site.admin_view(self.get_duplicate_season),
                           name='dr27-copy-season'),
                   ] + urls
        return new_urls


class RaceAdmin(CommonTabbedModelAdmin):
    list_display = ('__str__', 'season', 'print_pole', 'print_winner', 'print_fastest',)
    list_filter = ('season', 'season__competition', 'circuit', 'grand_prix',)

    form = RaceAdminForm
    tab_overview = (
        (None, {
            'fields': ('season', 'round', 'grand_prix', 'circuit', 'date', 'alter_punctuation', 'fastest_car')
        }),
    )
    tab_results = (
        ResultInline,
    )
    tabs = [
        ('Overview', tab_overview),
        # ('Summary', tab_summary),
        # ('Drivers', tab_drivers),
        ('Results', tab_results),
    ]

    # def get_urls(self):
    #     urls = super(RaceAdmin, self).get_urls()
    #     urlpatterns = [
    #         url(r'(?P<race_id>\d+)/results/$', self.admin_site.admin_view(self.results), name='driver27_race_results')
    #     ]
    #
    #     return urlpatterns + urls

    def print_seat(self, seat):
        return u"{driver}".format(driver=seat.driver) if seat else None

    def print_pole(self, obj):
        return self.print_seat(obj.pole)

    print_pole.short_description = _('Pole')

    def print_winner(self, obj):
        return self.print_seat(obj.winner)

    print_winner.short_description = _('Winner')

    def print_fastest(self, obj):
        return self.print_seat(obj.fastest)

    print_fastest.short_description = _('Fastest')

    @staticmethod
    def clean_position(position):
        try:
            position = int(position)
        except (ValueError, TypeError):
            position = None
        return position

    class Media:
        js = ['driver27/js/list_filter_collapse.js']


class SeatPeriodFilter(SimpleListFilter):
    title = 'period'
    parameter_name = 'period'

    def lookups(self, request, model_admin):
        periods = Season.objects.order_by().values_list('year', flat=True).distinct()
        return [(p, p) for p in periods]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(
                Q(periods__from_year__lte=self.value()) | Q(periods__from_year__isnull=True),
                Q(periods__until_year__gte=self.value()) | Q(periods__until_year__isnull=True)
            ).distinct()
        else:
            return queryset


class SeatAdmin(CommonTabbedModelAdmin):
    list_display = ('driver', 'team', 'print_periods')
    list_filter = (SeatPeriodFilter, 'driver', 'team', 'team__competitions')
    tab_overview = (
        (None, {
            'fields': ('team', 'driver')
        }),
    )
    tab_periods = (
        SeatPeriodInline,
    )
    tabs = [
        ('Overview', tab_overview),
        ('Periods', tab_periods),
    ]

    def print_periods(self, obj):
        periods = obj.periods.all()
        if periods.count():
            return ', '.join(['{from_year} to {until_year}'.format(from_year=period.from_year,
                                                                   until_year=period.until_year) for period in periods])
        else:
            return None

    print_periods.short_description = _('periods')

    class Media:
        js = ['driver27/js/list_filter_collapse.js']


class SeatPeriodAdmin(admin.ModelAdmin):
    list_filter = ('seat__driver', 'seat__team', 'seat__team__competitions',)

    class Media:
        js = ['driver27/js/list_filter_collapse.js']


admin.site.register(Driver, DriverAdmin)
admin.site.register(Team, TeamAdmin)
admin.site.register(Competition, CompetitionAdmin)
admin.site.register(Circuit, CircuitAdmin)
admin.site.register(GrandPrix, GrandPrixAdmin)
admin.site.register(Season, SeasonAdmin)
admin.site.register(Race, RaceAdmin)
admin.site.register(Seat, SeatAdmin)
admin.site.register(SeatPeriod, SeatPeriodAdmin)



from django import forms
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext as _
from tabbed_admin import TabbedModelAdmin

from django.utils.safestring import mark_safe
from django.utils.html import format_html
from django.utils.encoding import force_text


def get_circuit_id_from_gp(grand_prix_id):
    data_circuit_attr = ''
    if grand_prix_id:
        from driver27.models import GrandPrix
        grand_prix = GrandPrix.objects.filter(pk=grand_prix_id)
        if grand_prix.count() and grand_prix.first().default_circuit:
            data_circuit_attr = getattr(grand_prix.first().default_circuit, 'pk', '')
    return data_circuit_attr


class GrandPrixWidget(forms.widgets.Select):

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        # Add default_circuit as data to automatize circuit selection (>2.7)
        option = None
        if hasattr(super(GrandPrixWidget, self), 'create_option'):
            option = getattr(super(GrandPrixWidget, self), 'create_option')(name, value, label, selected, index, subindex, attrs)
            option['attrs']['data-circuit'] = get_circuit_id_from_gp(value)
        return option

    def render_option(self, selected_choices, option_value, option_label):
        # Add default_circuit as data to automatize circuit selection (==2.7)
        # Override method
        if option_value is None:
            option_value = ''
        option_value = force_text(option_value)
        if option_value in selected_choices:
            selected_html = mark_safe(' selected="selected"')
            if not self.allow_multiple_selected:
                # Only allow for a single selection.
                selected_choices.remove(option_value)
        else:
            selected_html = ''
        data_circuit_attr = get_circuit_id_from_gp(option_value)

        return format_html(u'<option value="{}"{} data-circuit="{}">{}</option>',
                           option_value, selected_html,
                           data_circuit_attr, force_text(option_label))

    class Media:
        js = ['driver27/js/select_default_circuit.js']





# http://stackoverflow.com/a/34567383
class AlwaysChangedModelForm(forms.ModelForm):
    def is_empty_form(self, *args, **kwargs):
        empty_form = True
        for name, field in self.fields.items():
            prefixed_name = self.add_prefix(name)
            data_value = field.widget.value_from_datadict(self.data, self.files, prefixed_name)
            if data_value:
                empty_form = False
                break
        return empty_form

    def has_changed(self, *args, **kwargs):
        """ Should returns True if data differs from initial.
        By always returning true even unchanged inlines will get validated and saved."""
        if self.instance.pk is None and self.initial:
            if not self.changed_data:
                return True
            if self.is_empty_form():
                return False
        return super(AlwaysChangedModelForm, self).has_changed()


class RelatedCompetitionAdmin(object):
    """ Aux class to share print_competitions method between driver and team """
    def print_competitions(self, obj):
        if hasattr(obj, 'competitions'):
            return ', '.join('{competition}'.format(competition=competition)
                             for competition in obj.competitions.all())
        else:
            return None
    print_competitions.short_description = _('competitions')


class CommonTabbedModelAdmin(TabbedModelAdmin):
    def get_form(self, request, obj=None, **kwargs):
        # just save obj reference for future processing in Inline
        if request and obj:
            request._obj_ = obj
        return super(CommonTabbedModelAdmin, self).get_form(request=request, obj=obj, **kwargs)



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from .common import retro_encode, CommonDriverTestCase
from ..models import Driver


class DriverTestCase(TestCase, CommonDriverTestCase):
    def test_driver_unicode(self):
        driver = self.get_test_driver()
        expected_unicode = ', '.join((driver.last_name, driver.first_name))
        self.assertEquals(str(driver), retro_encode(expected_unicode))


# -*- coding: utf-8 -*-
from django.test import TestCase
from .common import CommonTeamTestCase
from ..models import Team


class TeamTestCase(TestCase, CommonTeamTestCase):
    def test_team_unicode(self):
        team = self.get_test_team()
        self.assertTrue(str(team), team.name)


# encoding: utf-8
from .punctuation import get_punctuation_config
from django.db import models
from django.core.cache import cache
from six import text_type

try:
    from .models import Race
except ImportError:
    pass

try:
    from .models import Team
except ImportError:
    pass

from django.db.models import F, Sum
from .records import get_record_config


def order_points(rank):
    return sorted(rank, key=lambda x: (x['points'], x['pos_str']), reverse=True)


import re


class AbstractRankModel(models.Model):
    def get_stats_cls(self, contender):
        raise NotImplementedError('Not implemented method')

    def get_team_stats_cls(self, team):
        raise NotImplementedError('Not implemented method')


    @property
    def stats_filter_kwargs(self):
        raise NotImplementedError('Not implemented property')

    def get_name_cache_rank(self, prefix, kwargs):
        repr_kw = [u'{key}:{value}'.format(key=kw, value=kwargs.get(kw)) for kw in kwargs]
        repr_local = re.sub(r'[^\w:]', '', text_type(','.join(repr_kw)))
        cache_str = u'{prefix}_{repr_local}'.format(prefix=prefix, repr_local=repr_local)
        return cache_str

    def _abstract_points_rank_by_season(self, element_name, element_group, punctuation_config=None):
        rank = []
        items = getattr(self, element_group).all()
        for item in items:
            points_by_seasons = item.get_points_by_seasons(append_to_summary={element_name: item},
                                                             punctuation_config=punctuation_config,
                                                             **self.stats_filter_kwargs)
            for points_by_season in points_by_seasons:
                rank.append(points_by_season)
        return rank

    def _abstract_points_rank(self, element_name, element_group, stats_cls, punctuation_config=None):
        rank = []
        items= getattr(self, element_group).all()
        for item in items:
            stat_cls = stats_cls(item)
            rank.append(stat_cls.get_summary_points_rank(punctuation_config=punctuation_config,
                                                    exclude_position=True,
                                                    append_to_summary={element_name: item},
                                                    **self.stats_filter_kwargs))
        return rank

    def points_rank_by_season(self, punctuation_config=None):
        return self._abstract_points_rank_by_season('driver', 'drivers', punctuation_config)

    def team_points_rank_by_season(self, punctuation_config=None):
        return self._abstract_points_rank_by_season('team', 'teams', punctuation_config)

    def _points_rank(self, punctuation_config=None):
        return self._abstract_points_rank('driver', 'drivers', self.get_stats_cls, punctuation_config)

    def _team_points_rank(self, punctuation_config=None):
        return self._abstract_points_rank('team', 'teams', self.get_team_stats_cls, punctuation_config)

    def points_rank(self, punctuation_code=None, by_season=False):
        """ Points driver rank. Scoring can be override by scoring_code param """
        cache_str = self.get_name_cache_rank('points', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            punctuation_config = get_punctuation_config(punctuation_code=punctuation_code) \
                if punctuation_code is not None else None
            if by_season:
                rank = self.points_rank_by_season(punctuation_config=punctuation_config)
            else:
                rank = self._points_rank(punctuation_config=punctuation_config)
            cache.set(cache_str, rank)
        rank = order_points(rank)
        return rank

    def team_points_rank(self, punctuation_code=None, by_season=False):
        """ Same that points_rank by count both team drivers """
        cache_str = self.get_name_cache_rank('team_points', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            punctuation_config = get_punctuation_config(punctuation_code=punctuation_code) \
                if punctuation_code is not None else None
            teams = getattr(self, 'teams').all()
            if by_season:
                rank = self.team_points_rank_by_season(punctuation_config=punctuation_config)
            else:
                rank = self._team_points_rank(punctuation_config=punctuation_config)
            rank = order_points(rank)
            cache.set(cache_str, rank)
        return rank

    def olympic_rank(self):
        """ The driver
        with superior race results (based on descending order, from number of
        wins to numbers of second-places down) will gain precedence. """

        """ Points driver rank. Scoring can be override by scoring_code param """
        cache_str = self.get_name_cache_rank('olympic', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            drivers = getattr(self, 'drivers').all()
            rank = [self.get_stats_cls(driver).get_summary_points_rank(
                append_to_summary={'driver': driver},
                **self.stats_filter_kwargs) for driver in drivers]
        rank = sorted(rank, key=lambda x: x['pos_str'], reverse=True)
        cache.get(cache_str, rank)
        return rank

    def stats_rank(self, **filters):
        """ Get driver rank based on record filter """
        cache_str = self.get_name_cache_rank('stats', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            filters.update(**self.stats_filter_kwargs)
            drivers = getattr(self, 'drivers').all()
            rank = []
            for driver in drivers:
                stat_cls = self.get_stats_cls(driver)
                rank.append({'stat': stat_cls.get_stats(**filters),
                             'driver': driver,
                             'teams': stat_cls.teams_verbose})
            rank = sorted(rank, key=lambda x: x['stat'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    def comeback_rank(self):
        from .models import Result
        comeback_filter = get_record_config('COMEBACK').get('filter')
        comeback_filter.update(self.stats_filter_kwargs)
        cache_str = self.get_name_cache_rank('comeback', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            rank = Result.wizard(**comeback_filter) \
                .annotate(comeback=Sum(F('qualifying') - F('finish'))).order_by('-comeback')
            cache.set(cache_str, rank)
        return rank

    def seasons_rank(self, **filters):
        """ Get driver rank based on record filter """
        filters.update(**self.stats_filter_kwargs)
        cache_str = self.get_name_cache_rank('seasons', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            drivers = getattr(self, 'drivers').all()
            rank = []
            for driver in drivers:
                stat_cls = self.get_stats_cls(driver)
                driver_seasons = driver.seasons.filter(**self.stats_filter_kwargs)
                for season in driver_seasons:
                    rank.append({'stat': stat_cls.get_stats(season=season, **filters),
                                 'driver': driver,
                                 'teams': stat_cls.teams_verbose,
                                 'season': season})
            rank = sorted(rank, key=lambda x: x['stat'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    def seasons_team_rank(self, **filters):
        """ Get driver rank based on record filter """
        filters.update(**self.stats_filter_kwargs)
        cache_str = self.get_name_cache_rank('seasons_team', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            teams = getattr(self, 'teams').all()
            rank = []
            for team in teams:
                stat_cls = self.get_stats_cls(team)
                team_seasons = team.seasons.filter(**self.stats_filter_kwargs)
                for season in team_seasons:
                    rank.append({'stat': stat_cls.get_stats(season=season, **filters),
                                 'team': team,
                                 'season': season})
            rank = sorted(rank, key=lambda x: x['stat'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    def streak_rank(self, only_actives=False, max_streak=False, **filters):
        """ Get driver rank based on record filter """
        cache_str = self.get_name_cache_rank('streak', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            drivers = getattr(self, 'drivers').all()
            rank = []
            for driver in drivers:
                stat_cls = self.get_stats_cls(driver)
                if only_actives and not stat_cls.is_active:
                    continue
                rank.append({'stat': stat_cls.get_streak(result_filter=self.stats_filter_kwargs,
                                                         max_streak=max_streak, **filters),
                             'driver': driver,
                             'teams': stat_cls.teams_verbose})
            rank = sorted(rank, key=lambda x: x['stat'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    def streak_team_rank(self, only_actives=False, max_streak=False, **filters):
        """ Get team rank based on record filter """
        cache_str = self.get_name_cache_rank('streak_team', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            teams = getattr(self, 'teams').all()
            rank = []
            for team in teams:
                stat_cls = self.get_stats_cls(team)
                # if only_actives and not stat_cls.is_active:
                #     continue

                rank.append({'stat': stat_cls.get_streak(unique_by_race=True, result_filter=self.stats_filter_kwargs,
                                                         max_streak=max_streak, **filters),
                             'team': team})
            rank = sorted(rank, key=lambda x: x['stat'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    @staticmethod
    def get_team_rank_method(rank_type):
        """ Dict with distinct teams rank method, depending of rank_type param """
        " STATS: Count 1 by each time that driver get record "
        " RACES: Count 1 by each race that any driver get record "
        " DOUBLES: Count 1 by each race that at least two drivers get record "
        rank_dict = {
            'STATS': 'team_stats_rank',
            'RACES': 'team_races_rank',
            'DOUBLES': 'team_doubles_rank'
        }

        return rank_dict.get(rank_type)

    def get_team_rank(self, rank_type, **filters):
        """ Return team rank calling returned method by get_team_rank_method """
        rank_method = self.get_team_rank_method(rank_type)
        return getattr(self, rank_method)(**filters) if rank_method else None

        # if isinstance(self, Season):
        #     return TeamSeason.objects.get(season=self, team=team)
        # elif isinstance(self, Competition):
        #     return team

    def team_rank(self, total_method, **filters):
        """ Collect the records of each team calling the method of Team passed by total_method param """
        cache_str = self.get_name_cache_rank('team', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            rank = []
            teams = getattr(self, 'teams').all()
            for team in teams:
                team_stats_cls = self.get_team_stats_cls(team)
                total = getattr(team_stats_cls, total_method)(**filters)
                rank.append({'stat': total, 'team': team})
            rank = sorted(rank, key=lambda x: x['stat'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    def team_stats_rank(self, **filters):
        return self.team_rank('get_total_stats', **filters)

    def team_races_rank(self, **filters):
        return self.team_rank('get_total_races', **filters)

    def team_doubles_rank(self, **filters):
        return self.team_rank('get_doubles_races', **filters)

    def team_olympic_rank(self):
        """ Points team rank. Scoring can be override by scoring_code param """
        cache_str = self.get_name_cache_rank('olympic_team', locals())
        cache_rank = cache.get(cache_str)

        if cache_rank:
            rank = cache_rank
        else:
            teams = getattr(self, 'teams').all()
            rank = []
            for team in teams:
                stat_cls = self.get_stats_cls(team)
                position_count_list = stat_cls.get_positions_count_list(**self.stats_filter_kwargs)
                position_count_str = stat_cls.get_positions_count_str(position_list=position_count_list)
                rank.append({'pos_str': position_count_str,
                             'team': team,
                             'pos_list': position_count_list
                             })
            rank = sorted(rank, key=lambda x: x['pos_str'], reverse=True)
            cache.set(cache_str, rank)
        return rank

    class Meta:
        abstract = True



from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django.test import TestCase, Client, RequestFactory
from ..models import Season

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


class MockRequest(object):
    pass


class MockSuperUser(object):
    def has_perm(self, perm):
        return True


def get_request():
    request = MockRequest()
    request.user = MockSuperUser()
    return request


def get_fixtures_test():
    # django test uses a new db, while pytest use transaction no_commit in the same settings db
    if hasattr(settings, 'PYTEST_SETTING') and settings.PYTEST_SETTING:
        return None
    else:
        return ['driver27.json']


class FixturesTest(TestCase):
    fixtures = get_fixtures_test()


class ViewTest(FixturesTest):

    def setUp(self):
        self.site = AdminSite()
        self.client = Client()
        self.factory = RequestFactory()

    # def test_spider_admin(self):
    #     client = Client()
    #     client.login(username='admin', password='pass')
    #     base_admin = '/admin/driver27'
    #     models = ['circuit', 'competition', 'driver', 'grandprix', 'race', 'season', 'seat', 'team']
    #     for model in models:
    #         url = base_admin + '/' + model + '/'
    #         resp = client.get(url, follow=True)
    #         self.assertEqual(resp.status_code, 200, url + ' code='+str(resp.status_code))
    #         url = base_admin + '/' + model + '/add/'
    #         resp = client.get(url, follow=True)
    #         self.assertEqual(resp.status_code, 200, url + ' code='+str(resp.status_code))
    #         url = base_admin + '/' + model + '/1/change/'
    #         resp = client.get(url, follow=True)
    #         self.assertEqual(resp.status_code, 200, url + ' code='+str(resp.status_code))




from django.utils.timezone import datetime  # important if using timezones
from rest_framework.decorators import detail_route
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from .common import DR27ViewSet
from .serializers import RaceSerializer, ResultSerializer, SeatSerializer, SeasonSerializer
from .serializers import CircuitSerializer, GrandPrixSerializer, CompetitionSerializer
from .serializers import TeamSerializer, DriverSerializer, SeatPeriodSerializer
from .common import get_dict_from_rank_entry, get_dict_from_team_rank_entry, get_dict_from_races
from ..models import Competition, Driver, Race, Result, Season, Seat, Team, GrandPrix, Circuit, SeatPeriod
from django.db.models import Q
from ..punctuation import get_punctuation_config

class ShortPagePagination(PageNumberPagination):
    page_size = 20
    page_size_query_param = 'page_size'


class CommonDetailViewSet(object):
    def get_common_detail_route(self, request, rel_model, serializer_cls, filters=None):
        obj = self.get_object()
        if filters is None:
            filters = {}
        self.queryset = getattr(obj, rel_model).filter(**filters)
        serializer = serializer_cls(instance=self.queryset, many=True, context={'request': request})
        return Response(serializer.data)

    def get_seat_detail_route(self, request, filter=None, exclude=None):
        """ Seat can not be directly accessed from Race """
        if filter is None:
            filter = {}
        if exclude is None:
            exclude = {}

        self.queryset = Seat.objects.filter(**filter).exclude(**exclude)
        serializer = SeatSerializer(instance=self.queryset, many=True, context={'request': request})
        return Response(serializer.data)

class AbstractCommonSeatViewSet(object):
    @detail_route(methods=['get'], url_path='no-seats')
    def no_seats(self, request, pk=None):
        """ Seats with team in season:competition but not active in season """
        return getattr(self, 'get_common_detail_route')(request, 'no_seats', SeatSerializer)


# ViewSets define the view behavior.
class RaceViewSet(DR27ViewSet, CommonDetailViewSet, AbstractCommonSeatViewSet):
    queryset = Race.objects.all()
    serializer_class = RaceSerializer

    @detail_route(methods=['get'])
    def results(self, request, pk=None):
        return self.get_common_detail_route(request, 'results', ResultSerializer)

    @detail_route(methods=['get'])
    def seats(self, request, pk=None):
        obj = self.get_object()
        return self.get_seat_detail_route(request, filter={'results__race': obj})

    @detail_route(methods=['get'], url_path='no-start-seats')
    def no_start_seats(self, request, pk=None):
        """ Seats in this season that are not part of the race """

        return self.get_common_detail_route(request, 'no_seats', SeatSerializer)




class DR27CommonCompetitionViewSet(DR27ViewSet):
    @detail_route(methods=['get'], url_path='next-race')
    def next_race(self, request, pk=None):
        obj = self.get_object()
        queryset_params = {'date__gte': datetime.today()}
        if isinstance(obj, Season):
            self.queryset = obj.races.filter(**queryset_params).first()
        elif isinstance(obj, Competition):
            queryset_params['season__competition'] = obj
            self.queryset = Race.objects.filter(**queryset_params).first()
        else:
            raise NotImplementedError('Not Implemented Method')
        serializer = RaceSerializer(instance=self.queryset, many=False, context={'request': request})
        return Response(serializer.data)


# ViewSets define the view behavior.
class SeasonViewSet(DR27CommonCompetitionViewSet, CommonDetailViewSet, AbstractCommonSeatViewSet):
    queryset = Season.objects.all()
    serializer_class = SeasonSerializer

    @detail_route(methods=['get'])
    def races(self, request, pk=None):
        return self.get_common_detail_route(request, 'races', RaceSerializer)

    @detail_route(methods=['get'])
    def pending_races(self, request, pk=None):
        return self.get_common_detail_route(request, 'pending_races', RaceSerializer)

    @detail_route(methods=['get'])
    def past_races(self, request, pk=None):
        return self.get_common_detail_route(request, 'past_races', RaceSerializer)

    @detail_route(methods=['get'])
    def seats(self, request, pk=None):
        return self.get_common_detail_route(request, 'seats', SeatSerializer)


    @detail_route(methods=['get'])
    def teams(self, request, pk=None):
        return self.get_common_detail_route(request, 'teams', TeamSerializer)



    @detail_route(methods=['get'])
    def summary(self, request, pk=None):
        season = self.get_object()
        serializer_rank = {
            'id': season.pk,
            'competition_id': season.competition_id,
            'competition_name': season.competition.full_name,
            'year': season.year,
            'pending_points': season.pending_points(),
            'has_champion': season.has_champion(),
            'leader': get_dict_from_rank_entry(season.leader),
            'punctuation': season.punctuation,
            'punctuation_config': get_punctuation_config(season.punctuation),
            'num_pending_races': season.num_pending_races,
            'past_races_summary': get_dict_from_races(season.past_races.all()),
            'pending_races_summary': get_dict_from_races(season.pending_races.all())
        }
        return Response(serializer_rank)

    def _abstract_rank(self, rank_method, dict_method):
        season = self.get_object()
        rank = getattr(season, rank_method)()
        serializer_rank = [
            dict_method(standing) for standing in rank
        ]
        return Response(serializer_rank)

    @detail_route(methods=['get'])
    def standings(self, request, pk=None):
        return self._abstract_rank('points_rank', get_dict_from_rank_entry)

    @detail_route(methods=['get'], url_path='standings-team')
    def standings_team(self, request, pk=None):
        return self._abstract_rank('team_points_rank', get_dict_from_team_rank_entry)

    @detail_route(methods=['get'])
    def title(self, request, pk=None):
        return self._abstract_rank('only_title_contenders', get_dict_from_rank_entry)


# ViewSets define the view behavior.
class CircuitViewSet(DR27ViewSet):
    queryset = Circuit.objects.all()
    serializer_class = CircuitSerializer


# ViewSets define the view behavior.
class GrandPrixViewSet(DR27ViewSet):
    queryset = GrandPrix.objects.all()
    serializer_class = GrandPrixSerializer


# ViewSets define the view behavior.
class CompetitionViewSet(DR27CommonCompetitionViewSet, CommonDetailViewSet):
    queryset = Competition.objects.all()
    serializer_class = CompetitionSerializer

    @detail_route(methods=['get'])
    def teams(self, request, pk=None):
        return self.get_common_detail_route(request, 'teams', TeamSerializer)


# ViewSets define the view behavior.
class ResultViewSet(DR27ViewSet):
    queryset = Result.objects.all()
    serializer_class = ResultSerializer
    pagination_class = ShortPagePagination


# ViewSets define the view behavior.
class DriverViewSet(DR27CommonCompetitionViewSet, CommonDetailViewSet):
    queryset = Driver.objects.all()
    serializer_class = DriverSerializer

    @detail_route(methods=['get'])
    def seats(self, request, pk=None):
        return self.get_common_detail_route(request, 'seats', SeatSerializer)


# ViewSets define the view behavior.
class TeamViewSet(DR27ViewSet):
    queryset = Team.objects.all()
    serializer_class = TeamSerializer


# ViewSets define the view behavior.
class SeatViewSet(DR27ViewSet, CommonDetailViewSet):
    queryset = Seat.objects.all()
    serializer_class = SeatSerializer

    @detail_route(methods=['get'])
    def periods(self, request, pk=None):
        return self.get_common_detail_route(request, 'periods', SeatPeriodSerializer)


# ViewSets define the view behavior.
class SeatPeriodViewSet(DR27ViewSet):
    queryset = SeatPeriod.objects.all()
    serializer_class = SeatPeriodSerializer



class PointsCalculator(object):
    def __init__(self, punctuation_config):
        self.punctuation_config = punctuation_config

    def get_points_qualifying(self, qualifying):
        qualifying_scoring = self.punctuation_config.get('qualifying')
        if qualifying_scoring and qualifying:
            scoring_len = len(qualifying_scoring)
            if not qualifying > scoring_len:
                return qualifying_scoring[qualifying - 1]
        return 0

    def get_points_finish(self, finish, alter_punctuation):
        points_factor = {'double': 2, 'half': 0.5}
        factor = points_factor[alter_punctuation] if alter_punctuation in points_factor else 1
        finish_scoring = self.punctuation_config.get('finish')
        if finish_scoring and finish:
            scoring_len = len(finish_scoring)
            if not finish > scoring_len:
                return finish_scoring[finish - 1] * factor
        return 0

    def get_points_fastest_lap(self, fastest_lap):
        if 'fastest_lap' in self.punctuation_config and fastest_lap:
            return self.punctuation_config.get('fastest_lap')
        else:
            return 0

    def calculator(self, result, skip_wildcard=False):
        points = 0
        if result.wildcard and skip_wildcard:
            return points
        points += self.get_points_qualifying(result.qualifying)
        points += self.get_points_finish(result.finish, result.alter_punctuation)
        points += self.get_points_fastest_lap(result.fastest_lap)
        return points



from django_countries.serializer_fields import CountryField
from rest_framework import serializers
from ..models import GrandPrix, Circuit, Season, Competition, Driver, Team, Seat, Race, Result, SeatPeriod
from .common import DR27Serializer
from ..common import DRIVER27_API_NAMESPACE, DRIVER27_NAMESPACE

def get_view_name_detail(detail):
    return DRIVER27_NAMESPACE + ':' + DRIVER27_API_NAMESPACE + ':' + detail + '-detail'

class GrandPrixSerializer(DR27Serializer, serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    country = CountryField()

    class Meta:
        model = GrandPrix
        fields = ('url', 'id', 'country', 'name', 'first_held', 'default_circuit', 'competitions',)
        extra_kwargs = {'url': {'view_name': get_view_name_detail('grandprix')}}


class CircuitSerializer(DR27Serializer, serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    country = CountryField()

    class Meta:
        model = Circuit
        fields = ('url', 'id', 'country', 'name', 'city', 'opened_in',)
        extra_kwargs = {'url': {'view_name': get_view_name_detail('circuit')}}


class SeasonSerializer(DR27Serializer, serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    competition_details = serializers.SerializerMethodField()
    slug = serializers.SerializerMethodField(read_only=True)
    races = serializers.HyperlinkedRelatedField(view_name=get_view_name_detail('race'),
                                                many=True,
                                                read_only=True)

    def get_competition_details(self, obj):
        return CompetitionSerializer(instance=obj.competition, many=False,
                                     context=self.context, exclude_fields=['seasons', ]).data

    def get_slug(self, obj):
        return '-'.join((obj.competition.slug, str(obj.year)))

    class Meta:
        model = Season
        fields = ('url', 'id', 'year', 'rounds', 'slug', 'punctuation', 'competition',
                  'competition_details', 'races',)
        read_only_fields = ('competition_details', 'races',)
        extra_kwargs = {'url': {'view_name': get_view_name_detail('season')}}


class CompetitionSerializer(DR27Serializer, serializers.HyperlinkedModelSerializer):
    id = serializers.ReadOnlyField()
    # https://github.com/SmileyChris/django-countries/issues/106
    country = CountryField()
    seasons = SeasonSerializer(many=True, exclude_fields=['competition', 'competition_details', 'races'],
                               read_only=True)

    class Meta:
        model = Competition
        fields = ('url', 'id', 'name', 'full_name', 'country', 'slug', 'seasons')
        extra_kwargs = {'url': {'view_name': get_view_name_detail('competition')}}


class DriverSerializer(serializers.HyperlinkedModelSerializer):
    id = serializers.ReadOnlyField()
    country = CountryField()

    class Meta:
        model = Driver
        fields = ('url', 'id', 'last_name', 'first_name', 'year_of_birth', 'country', 'seats')
        read_only_fields = ('seats',)
        extra_kwargs = {'url': {'view_name': get_view_name_detail('driver')},
                        'seats': {'view_name': get_view_name_detail('seat')}
                        }


class NestedDriverSerializer(DriverSerializer):
    class Meta:
        model = Driver
        fields = ('url', 'last_name', 'first_name', 'year_of_birth', 'country')


class TeamSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    country = CountryField()

    class Meta:
        model = Team
        fields = ('url', 'id', 'name', 'full_name', 'competitions', 'country')
        extra_kwargs = {'url': {'view_name': get_view_name_detail('team')}}


class NestedTeamSerializer(TeamSerializer):
    class Meta:
        model = Team
        fields = ('url', 'name', 'full_name', 'country')


class SeatPeriodSerializer(serializers.ModelSerializer):
    class Meta:
        model = SeatPeriod
        fields = '__all__'


class SeatSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    team_details = serializers.SerializerMethodField()
    driver_details = serializers.SerializerMethodField()
    periods = serializers.SerializerMethodField()

    def get_team_details(self, obj):
        return TeamSerializer(instance=obj.team, many=False,
                              context=self.context).data

    def get_driver_details(self, obj):
        return DriverSerializer(instance=obj.driver, many=False, context=self.context).data

    def get_periods(self, obj):
        return SeatPeriodSerializer(instance=obj.periods, many=True, context=self.context).data

    class Meta:
        model = Seat
        fields = ('url', 'id', 'team', 'team_details', 'driver', 'driver_details', 'periods')
        read_only_fields = ('periods',)
        extra_kwargs = {'url': {'view_name': get_view_name_detail('seat')}}


class SeatRecapSerializer(serializers.ModelSerializer):
    driver_details = serializers.SerializerMethodField(method_name='get_driver')
    team = serializers.SerializerMethodField()

    def get_driver(self, obj):
        driver = obj.driver
        return {
            'id': driver.id,
            'first_name': driver.first_name,
            'last_name': driver.last_name
        }

    def get_team(self, obj):
        team = obj.team
        return {
            'id': team.id,
            'name': team.name
        }

    class Meta:
        model = Seat
        fields = ('driver_details', 'team')


class ResultSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    seat_details = serializers.SerializerMethodField()
    fastest_lap = serializers.SerializerMethodField()

    def get_fastest_lap(self, obj):
        return obj.is_fastest

    def get_seat_details(self, obj):
        return SeatRecapSerializer(instance=obj.seat,
                                   many=False,
                                   context=self.context).data

    class Meta:
        model = Result
        fields = ('url',
                  'id',
                  'race',
                  'seat',
                  'seat_details',
                  'qualifying',
                  'finish',
                  'fastest_lap',
                  'wildcard',
                  'retired',
                  'comment')
        extra_kwargs = {'url': {'view_name': get_view_name_detail('result')}}


class RaceSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    grand_prix_details = serializers.SerializerMethodField()
    circuit_details = serializers.SerializerMethodField()

    def get_grand_prix_details(self, obj):
        return GrandPrixSerializer(instance=obj.grand_prix,
                                   many=False,
                                   context=self.context,
                                   exclude_fields=['competitions', ]).data

    def get_circuit_details(self, obj):
        return CircuitSerializer(instance=obj.circuit,
                                 many=False,
                                 context=self.context).data

    class Meta:
        model = Race
        fields = ('url', 'id', 'season', 'round', 'grand_prix', 'grand_prix_details',
                  'circuit', 'circuit_details', 'date', 'alter_punctuation')

        extra_kwargs = {'url': {'view_name': get_view_name_detail('race')}}



import os
from .settings import *

# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.sqlite3',
#         'NAME': os.path.join(BASE_DIR, 'pytest.sqlite3'),
#     }
# }

PYTEST_SETTING = True
ALLOWED_HOSTS.append('testserver')

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': '127.0.0.1:11211',
        'KEY_PREFIX': 'driver27-demo-test'

    }
}